#ifndef _ASM_X86_REFCOUNT_H
#define _ASM_X86_REFCOUNT_H

#ifndef CONFIG_PAX_REFCOUNT
#error PAX_REFCOUNT must be enabled!
#endif

static __always_inline __must_check bool refcount_add_not_zero(unsigned int i, refcount_t *r)
{
	atomic_t *v = &r->refs;
	const int a = i, u = 0;
	int c, old, new;

	c = atomic_read(v);
	for (;;) {
		if (unlikely(c == u))
			break;

		asm volatile("addl %2,%0\n\t"
			     __PAX_REFCOUNT(4, "s")
			     : "=r" (new)
			     : "0" (c), "ir" (a),
			       [counter] "m" (v->counter)
			     : "cc", "cx");

		old = atomic_cmpxchg(v, c, new);
		if (likely(old == c))
			break;
		c = old;
	}
	return c != u;
}

static __always_inline void refcount_add(unsigned int i, refcount_t *r)
{
	asm volatile(LOCK_PREFIX "addl %1,%0\n\t"
		__PAX_REFCOUNT(4, "s")
		: [counter] "+m" (r->refs.counter)
		: "ir" (i)
		: "cc", "cx");
}

static __always_inline __must_check bool refcount_inc_not_zero(refcount_t *r)
{
	return refcount_add_not_zero(1, r);
}

static __always_inline void __refcount_inc(refcount_t *r, int *oldp)
{
	int old = atomic_fetch_add_relaxed(1, &r->refs);

	if (oldp)
		*oldp = old;
}

static __always_inline void refcount_inc(refcount_t *r)
{
	asm volatile(LOCK_PREFIX "incl %0\n\t"
		__PAX_REFCOUNT(4, "s")
		: [counter] "+m" (r->refs.counter)
		: : "cc", "cx");
}

static __always_inline __must_check bool refcount_sub_and_test(unsigned int i, refcount_t *r)
{
	bool ret = GEN_BINARY_RMWcc_refcount(LOCK_PREFIX "subl", r->refs.counter, -4, e, "er", i, "cx");

	if (ret) {
		smp_acquire__after_ctrl_dep();
		return true;
	}

	return false;
}

static __always_inline void refcount_sub(unsigned int i, refcount_t *r)
{
	asm volatile(LOCK_PREFIX "subl %1,%0\n\t"
		     __PAX_REFCOUNT(-4, "s")
		     : [counter] "+m" (r->refs.counter)
		     : "ir" (i)
		     : "cc", "cx");
}

static __always_inline __must_check bool __refcount_dec_and_test(refcount_t *r, int *oldp)
{
	int old = atomic_fetch_sub_release(1, &r->refs);

	if (oldp)
		*oldp = old;

	if (old == 1) {
		smp_acquire__after_ctrl_dep();
		return true;
	}

	return false;
}

static __always_inline __must_check bool refcount_dec_and_test(refcount_t *r)
{
	bool ret = GEN_UNARY_RMWcc_refcount(LOCK_PREFIX "decl", r->refs.counter, -4, e, "cx");

	if (ret) {
		smp_acquire__after_ctrl_dep();
		return true;
	}

	return false;
}

static __always_inline void refcount_dec(refcount_t *r)
{
	asm volatile(LOCK_PREFIX "decl %0\n\t"
		     __PAX_REFCOUNT(-4, "s")
		     : [counter] "+m" (r->refs.counter)
		     : : "cc", "cx");
}

static __always_inline __must_check bool refcount_dec_if_one(refcount_t *r)
{
	int val = 1;

	return atomic_try_cmpxchg_release(&r->refs, &val, 0);
}

static __always_inline __must_check bool refcount_dec_not_one(refcount_t *r)
{
	unsigned int new, val = atomic_read(&r->refs);

	do {
		if (unlikely(val == 1))
			return false;

		asm volatile("subl %2,%0\n\t"
			     __PAX_REFCOUNT(4, "s")
			     : "=r" (new)
			     : "0" (val), "ir" (1),
			       [counter] "m" (r->refs.counter)
			     : "cc", "cx");
	} while (!atomic_try_cmpxchg_release(&r->refs, &val, new));

	return true;
}

#endif
