/* SPDX-License-Identifier: GPL-2.0 */
/* Copyright (C) 2008  PaX Team <pageexec@freemail.hu> */

#ifndef _ASM_X86_UACCESS_64_H
#define _ASM_X86_UACCESS_64_H

/*
 * User space memory access functions
 */
#include <linux/compiler.h>
#include <linux/lockdep.h>
#include <linux/kasan-checks.h>
#include <asm/alternative.h>
#include <asm/cpufeatures.h>
#include <asm/page.h>

/*
 * Copy To/From Userspace
 */

/* Handles exceptions in both to and from, but doesn't do access_ok */
__must_check unsigned long
copy_user_enhanced_fast_string(void *to, const void *from, unsigned long len);
__must_check unsigned long
copy_user_generic_string(void *to, const void *from, unsigned long len);
__must_check unsigned long
copy_user_generic_unrolled(void *to, const void *from, unsigned long len);

static __always_inline __must_check __size_overflow(3) unsigned long
copy_user_generic(void *to, const void *from, unsigned long len)
{
	unsigned long ret;

	/*
	 * If CPU has ERMS feature, use copy_user_enhanced_fast_string.
	 * Otherwise, if CPU has rep_good feature, use copy_user_generic_string.
	 * Otherwise, use copy_user_generic_unrolled.
	 */
	alternative_call_2(copy_user_generic_unrolled,
			 copy_user_generic_string,
			 X86_FEATURE_REP_GOOD,
			 copy_user_enhanced_fast_string,
			 X86_FEATURE_ERMS,
			 ASM_OUTPUT2("=a" (ret), "=D" (to), "=S" (from),
				     "=d" (len)),
			 "1" (to), "2" (from), "3" (len)
			 : "memory", "rcx", "r8", "r9", "r10", "r11");
	return ret;
}

static __always_inline __must_check unsigned long
raw_copy_from_user(void *dst, const void __user *src, unsigned long size)
{
	access_ok_nospec(src, size);

#ifdef CONFIG_PAX_MEMORY_UDEREF
	if (unlikely((unsigned long)src >= TASK_SIZE_MAX))
		return size;
	if (!access_ok(src, size))
		return size;
#endif

	return copy_user_generic(dst, (__force_kernel const void *)src, size);
}

static __always_inline __must_check unsigned long
raw_copy_to_user(void __user *dst, const void *src, unsigned long size)
{

#ifdef CONFIG_PAX_MEMORY_UDEREF
	if (unlikely((unsigned long)dst >= TASK_SIZE_MAX))
		return size;
	if (!access_ok(dst, size))
		return size;
#endif

	return copy_user_generic((__force_kernel void *)dst, src, size);
}

extern unsigned long __copy_user_nocache(void *dst, const void __user *src,
				unsigned long size, int zerorest) __size_overflow(3);

extern long __copy_user_flushcache(void *dst, const void __user *src, size_t size) __size_overflow(3);
extern void memcpy_page_flushcache(char *to, struct page *page, size_t offset,
			   size_t len);

static inline unsigned long
__copy_from_user_inatomic_nocache(void *dst, const void __user *src,
				  unsigned long size)
{
	kasan_check_write(dst, size);

	if (size > INT_MAX)
		return size;

	if (check_object_size(dst, size, false))
		return size;

#ifdef CONFIG_PAX_MEMORY_UDEREF
	if (!access_ok(src, size))
		return size;
#endif

	return __copy_user_nocache(dst, src, size, 0);
}

static inline int
__copy_from_user_flushcache(void *dst, const void __user *src, size_t size)
{
	kasan_check_write(dst, size);
	return __copy_user_flushcache(dst, src, size);
}
#endif /* _ASM_X86_UACCESS_64_H */
