/* SPDX-License-Identifier: GPL-2.0 */
#ifndef _ASM_X86_VDSO_H
#define _ASM_X86_VDSO_H

#include <asm/page_types.h>
#include <linux/linkage.h>
#include <linux/init.h>

#ifdef CONFIG_PAX_MEMORY_UDEREF
/* Bit used for the pseudo-hwcap for non-negative segments.  We use
   bit 1 to avoid bugs in some versions of glibc when bit 0 is
   used; the choice is otherwise arbitrary. */
#define VDSO_NOTE_NOSEGNEG_BIT	1

/* This is just a hack for compatibility with glibc 2.26+ which broke the
 * support for extra hwcaps by having ldconfig and ld.so divert on where these
 * should start -- ldconfig will use bit 52, ld.so uses bit 50. So by setting
 * bit 1 of the extra bits we will never be able to hit the bit ldconfig used
 * when creating /etc/ld.so.cache. We therefore use an additional bit 3 here
 * which will map to ldconfig's bit 1 above to work around that issue. But
 * beware, this will break again in the future, likely in glibc 2.32:
 *
 *   https://sourceware.org/git/?p=glibc.git;a=commitdiff;h=dcbc6b83eff5
 *   https://sourceware.org/git/?p=glibc.git;a=commitdiff;h=31563b68410f
 */
#define VDSO_NOTE_NOSEGNEG_BIT2	3
#endif

#ifndef __ASSEMBLER__

#include <linux/mm_types.h>

struct vdso_image {
	void *data;
	unsigned long size;   /* Always a multiple of PAGE_SIZE */

	unsigned long alt, alt_len;
	unsigned long extable_base, extable_len;
	const void *extable;

	long sym_vvar_start;  /* Negative offset to the vvar area */

	long sym_vvar_page;
	long sym_pvclock_page;
	long sym_hvclock_page;
	long sym_timens_page;
	long sym_VDSO32_NOTE_MASK;
	long sym___kernel_sigreturn;
	long sym___kernel_rt_sigreturn;
	long sym___kernel_vsyscall;
	long sym_int80_landing_pad;
	long sym_vdso32_sigreturn_landing_pad;
	long sym_vdso32_rt_sigreturn_landing_pad;
};

#ifdef CONFIG_X86_64
extern const struct vdso_image vdso_image_64;
#endif

#ifdef CONFIG_X86_X32
extern const struct vdso_image vdso_image_x32;
#endif

#if defined CONFIG_X86_32 || defined CONFIG_COMPAT
extern const struct vdso_image vdso_image_32;
#endif

extern void __init init_vdso_image(const struct vdso_image *image);

extern int map_vdso_once(const struct vdso_image *image, unsigned long addr);

extern bool fixup_vdso_exception(struct pt_regs *regs, int trapnr,
				 unsigned long error_code,
				 unsigned long fault_addr);
#endif /* __ASSEMBLER__ */

#endif /* _ASM_X86_VDSO_H */
