/* SPDX-License-Identifier: GPL-2.0 */
#ifndef __LINUX_COMPILER_TYPES_H
#define __LINUX_COMPILER_TYPES_H

#ifndef __ASSEMBLY__

#ifdef __CHECKER__
/* address spaces */
# define __kernel	__attribute__((address_space(0)))
# define __force_kernel	__force __kernel
# define __user		__attribute__((noderef, address_space(__user)))
# define __force_user	__force __user
# define __iomem	__attribute__((noderef, address_space(__iomem)))
# define __force_iomem	__force __iomem
# define __percpu	__attribute__((noderef, address_space(__percpu)))
# define __force_percpu	__force __percpu
# define __rcu		__attribute__((noderef, address_space(__rcu)))
# define __force_rcu	__force __rcu
static inline void __chk_user_ptr(const volatile void __user *ptr) { }
static inline void __chk_io_ptr(const volatile void __iomem *ptr) { }
/* context/locking */
# define __must_hold(x)	__attribute__((context(x,1,1)))
# define __acquires(x)	__attribute__((context(x,0,1)))
# define __releases(x)	__attribute__((context(x,1,0)))
# define __acquire(x)	__context__(x,1)
# define __release(x)	__context__(x,-1)
# define __cond_lock(x,c)	((c) ? ({ __acquire(x); 1; }) : 0)
# define __opens(x)	__attribute__((context(#x,0,1)))
# define __closes(x)	__attribute__((context(#x,1,0)))
# define __open(x)	__context__(#x,1)
# define __close(x)	__context__(#x,-1)
/* other */
# define __force	__attribute__((force))
# define __nocast	__attribute__((nocast))
# define __safe		__attribute__((safe))
# define __private	__attribute__((noderef))
# define ACCESS_PRIVATE(p, member) (*((typeof((p)->member) __force *) &(p)->member))
#else /* __CHECKER__ */
/* address spaces */
# ifdef CHECKER_PLUGIN
#  ifdef CHECKER_PLUGIN_USER
//#  define __user
//#  define __force_user
//#  define __kernel
//#  define __force_kernel
#  else /* CHECKER_PLUGIN_USER */
#   define __user
#   define __force_user
#   define __kernel
#   define __force_kernel
#  endif /* CHECKER_PLUGIN_USER */
#  ifdef CHECKER_PLUGIN_CONTEXT
/* context/locking */
#   define __must_hold(x)	__attribute__((context(#x,1,1)))
#   define __acquires(x)	__attribute__((context(#x,0,1)))
#   define __releases(x)	__attribute__((context(#x,1,0)))
#   define __acquire(x)		__context__(#x,1)
#   define __release(x)		__context__(#x,-1)
#   define __cond_lock(x,c)	((c) ? ({ __acquire(x); 1; }) : 0)
#   define __cond_unlock(x,c)	((c) ? ({ __release(x); 1; }) : 0)
#  else /* CHECKER_PLUGIN_CONTEXT */
/* context/locking */
#   define __must_hold(x)
#   define __acquires(x)
#   define __releases(x)
#   define __acquire(x) (void)0
#   define __release(x) (void)0
#   define __cond_lock(x,c) (c)
#   define __cond_unlock(x,c) (c)
#  endif /* CHECKER_PLUGIN_CONTEXT */
# else /* CHECKER_PLUGIN */
#  ifdef STRUCTLEAK_PLUGIN
#   define __user __attribute__((user))
#  else
#   define __user
#  endif
#  define __force_user
#  define __kernel
#  define __force_kernel
#  if 1
/* context/locking */
#   define __must_hold(x)
#   define __acquires(x)
#   define __releases(x)
#   define __acquire(x) (void)0
#   define __release(x) (void)0
#   define __cond_lock(x,c) (c)
#  else
/* context/locking */
#   define __must_hold(x)	__attribute__((context(#x,1,1)))
#   define __acquires(x)	__attribute__((context(#x,0,1)))
#   define __releases(x)	__attribute__((context(#x,1,0)))
#   define __acquire(x)		__context__(#x,1)
#   define __release(x)		__context__(#x,-1)
#   define __cond_lock(x,c)	((c) ? ({ __acquire(x); 1; }) : 0)
#   define __cond_unlock(x,c)	((c) ? ({ __release(x); 1; }) : 0)
# endif
# endif /* CHECKER_PLUGIN */
# define __iomem
# define __force_iomem
# define __percpu
# define __force_percpu
# define __rcu
# define __force_rcu
# define __chk_user_ptr(x)	(void)0
# define __chk_io_ptr(x)	(void)0
/* other */
# define __force
# define __nocast
# define __safe
# define __private
# define ACCESS_PRIVATE(p, member) ((p)->member)
# define __builtin_warning(x, y...) (1)
# ifdef CONSTIFY_PLUGIN
/* context/locking */
#  define __opens(x)	__attribute__((context(#x,0,1)))
#  define __closes(x)	__attribute__((context(#x,1,0)))
#  define __open(x)	__context__(#x,1)
#  define __close(x)	__context__(#x,-1)
# else /* CONSTIFY_PLUGIN */
/* context/locking */
#  define __opens(x)
#  define __closes(x)
#  define __open(x)	(void)0
#  define __close(x)	(void)0
# endif /* CONSTIFY_PLUGIN */
#endif /* __CHECKER__ */

#ifdef __KERNEL__

/* Attributes */
#include <linux/compiler_attributes.h>

/* Builtins */

/*
 * __has_builtin is supported on gcc >= 10, clang >= 3 and icc >= 21.
 * In the meantime, to support gcc < 10, we implement __has_builtin
 * by hand.
 */
#ifndef __has_builtin
#define __has_builtin(x) (0)
#endif

/*
 * Some architectures need to provide custom definitions of macros provided
 * by linux/compiler-*.h, and can do so using asm/compiler.h. We include that
 * conditionally rather than using an asm-generic wrapper in order to avoid
 * build failures if any C compilation, which will include this file via an
 * -include argument in c_flags, occurs prior to the asm-generic wrappers being
 * generated.
 */
#ifdef CONFIG_HAVE_ARCH_COMPILER_H
#include <asm/compiler.h>
#endif

#ifndef RAP_SAFE_ASM
#define RAP_SAFE_ASM
#endif

/* Compiler specific macros. */
#ifdef __clang__
#include <linux/compiler-clang.h>
#elif defined(__INTEL_COMPILER)
#include <linux/compiler-intel.h>
#elif defined(__GNUC__)
/* The above compilers also define __GNUC__, so order is important here. */
#include <linux/compiler-gcc.h>
#else
#error "Unknown compiler"
#endif

struct ftrace_branch_data {
	const char *func;
	const char *file;
	unsigned line;
	union {
		struct {
			unsigned long correct;
			unsigned long incorrect;
		};
		struct {
			unsigned long miss;
			unsigned long hit;
		};
		unsigned long miss_hit[2];
	};
};

struct ftrace_likely_data {
	struct ftrace_branch_data	data;
	unsigned long			constant;
};

#if defined(CC_USING_HOTPATCH)
#define notrace			__attribute__((hotpatch(0, 0)))
#elif defined(CC_USING_PATCHABLE_FUNCTION_ENTRY)
#define notrace			__attribute__((patchable_function_entry(0, 0)))
#else
#define notrace			__attribute__((__no_instrument_function__))
#endif

/*
 * it doesn't make sense on ARM (currently the only user of __naked)
 * to trace naked functions because then mcount is called without
 * stack and frame pointer being set up and there is no chance to
 * restore the lr register to the value before mcount was called.
 */
#if defined(CONFIG_X86) && defined(CONFIG_CC_IS_GCC)
# if CONFIG_GCC_VERSION >= 60000
#  define __naked			__attribute__((__naked__)) notrace
# else
#  define __naked			notrace
# endif
#else
# define __naked			__attribute__((__naked__)) notrace
#endif

#define __compiler_offsetof(a, b)	__builtin_offsetof(a, b)

/*
 * Prefer gnu_inline, so that extern inline functions do not emit an
 * externally visible function. This makes extern inline behave as per gnu89
 * semantics rather than c99. This prevents multiple symbol definition errors
 * of extern inline functions at link time.
 * A lot of inline functions can cause havoc with function tracing.
 */
#define inline inline __gnu_inline __inline_maybe_unused notrace

/*
 * gcc provides both __inline__ and __inline as alternate spellings of
 * the inline keyword, though the latter is undocumented. New kernel
 * code should only use the inline spelling, but some existing code
 * uses __inline__. Since we #define inline above, to ensure
 * __inline__ has the same semantics, we need this #define.
 *
 * However, the spelling __inline is strictly reserved for referring
 * to the bare keyword.
 */
#define __inline__ inline

/*
 * GCC does not warn about unused static inline functions for -Wunused-function.
 * Suppress the warning in clang as well by using __maybe_unused, but enable it
 * for W=1 build. This will allow clang to find unused functions. Remove the
 * __inline_maybe_unused entirely after fixing most of -Wunused-function warnings.
 */
#ifdef KBUILD_EXTRA_WARN1
#define __inline_maybe_unused
#else
#define __inline_maybe_unused __maybe_unused
#endif

/*
 * Rather then using noinline to prevent stack consumption, use
 * noinline_for_stack instead.  For documentation reasons.
 */
#define noinline_for_stack noinline

/*
 * Sanitizer helper attributes: Because using __always_inline and
 * __no_sanitize_* conflict, provide helper attributes that will either expand
 * to __no_sanitize_* in compilation units where instrumentation is enabled
 * (__SANITIZE_*__), or __always_inline in compilation units without
 * instrumentation (__SANITIZE_*__ undefined).
 */
#ifdef __SANITIZE_ADDRESS__
/*
 * We can't declare function 'inline' because __no_sanitize_address conflicts
 * with inlining. Attempt to inline it may cause a build failure.
 *     https://gcc.gnu.org/bugzilla/show_bug.cgi?id=67368
 * '__maybe_unused' allows us to avoid defined-but-not-used warnings.
 */
# define __no_kasan_or_inline __no_sanitize_address notrace __maybe_unused
# define __no_sanitize_or_inline __no_kasan_or_inline
#else
# define __no_kasan_or_inline __always_inline
#endif

#define __no_kcsan __no_sanitize_thread
#ifdef __SANITIZE_THREAD__
# define __no_sanitize_or_inline __no_kcsan notrace __maybe_unused
#endif

#ifndef __no_sanitize_or_inline
#define __no_sanitize_or_inline __always_inline
#endif

/* Section for code which can't be instrumented at all */
#define noinstr								\
	noinline notrace __attribute((__section__(".noinstr.text")))	\
	__no_kcsan __no_sanitize_address __no_profile __no_sanitize_coverage

/* Disable inlining especially under LTO that would otherwise lead to autoslab name collisions */
#define noinline_for_autoslab noinline

#endif /* __KERNEL__ */

#endif /* __ASSEMBLY__ */

/* Indirect macros required for expanded argument pasting, eg. __LINE__. */
#define ___PASTE(a,b) a##b
#define __PASTE(a,b) ___PASTE(a,b)

/*
 * The below symbols may be defined for one or more, but not ALL, of the above
 * compilers. We don't consider that to be an error, so set them to nothing.
 * For example, some of them are for compiler specific plugins.
 */
#ifndef __latent_entropy
# define __latent_entropy
#endif

#ifndef __randomize_layout
# define __randomize_layout __designated_init
#endif

#ifndef __no_randomize_layout
# define __no_randomize_layout
#endif

#ifndef __no_const
# define __no_const
#endif

#ifndef __do_const
# define __do_const
#endif

#ifndef __mutable_const
# define __mutable_const
#endif

#ifndef __mutable
# define __mutable
#endif

#ifndef __size_overflow
# define __size_overflow(...)
#endif

#ifndef __intentional_overflow
# define __intentional_overflow(...)
# define __turn_off_size_overflow
# define __skip_size_overflow
#endif

#ifndef __nocapture
# define __nocapture(...)
#endif

#ifndef __noderef
# define __noderef
#endif

#ifndef __leaf_only
# define __leaf_only
#endif

#ifndef __unverified_nocapture
# define __unverified_nocapture(...)
#endif

#ifndef __rap_hash
#define __rap_hash
#endif

#ifndef __rap_hash_icall
#define __rap_hash_icall
#endif

#ifndef __rap_hash_override
#define __rap_hash_override(...)
#endif

#ifndef __respectre_inline
#define __respectre_inline __always_inline
#endif

#ifndef __read_once
#define __read_once
#endif

#ifndef __write_once
#define __write_once
#endif

#ifndef __section_type
#define __section_type(...)
#endif

#ifndef __maybe_used
#define __maybe_used
#endif

#ifndef __alloc_size
#define __alloc_size(...)
#endif

#ifndef __alloc_gfp_flags
#define __alloc_gfp_flags(...)
#endif

#ifndef __alloc_xfs_km_flags
#define __alloc_xfs_km_flags(...)
#endif

#ifndef __alloc_node
#define __alloc_node(...)
#endif

#ifndef __alloc
#define __alloc
#endif

#ifndef __bos
#define __bos(ptr, arg)
#endif

#ifndef __bos0
#define __bos0(ptr)
#endif

#ifndef __bos1
#define __bos1(ptr)
#endif

#ifndef __noscs
# define __noscs
#endif

#ifndef __nocfi
# define __nocfi
#endif

#ifndef __cficanonical
# define __cficanonical
#endif

#ifndef asm_volatile_goto
#define asm_volatile_goto(x...) asm goto(x)
#endif

#ifdef CONFIG_CC_HAS_ASM_INLINE
#define asm_inline asm __inline
#else
#define asm_inline asm
#endif

/* Are two types/vars the same type (ignoring qualifiers)? */
#define __same_type(a, b) __builtin_types_compatible_p(typeof(a), typeof(b))

#define __type_is_unsigned(t) (__same_type((t)0, 0UL) || __same_type((t)0, 0U) || __same_type((t)0, (unsigned short)0) || __same_type((t)0, (unsigned char)0))

#define ____inttypeof(t, v)	\
	__builtin_choose_expr(__builtin_types_compatible_p(typeof(v), t), (t)0,

#define __inttypeof(t, v)	\
	____inttypeof(signed t, v)\
	____inttypeof(unsigned t, v)

#define inttypeof(v)	typeof(\
	__inttypeof(char, v) \
	__inttypeof(short, v) \
	__inttypeof(int, v) \
	__inttypeof(long, v) \
	__inttypeof(long long, v) \
	(void)0)))))))))))

/*
 * __unqual_scalar_typeof(x) - Declare an unqualified scalar type, leaving
 *			       non-scalar types unchanged.
 */
/*
 * Prefer C11 _Generic for better compile-times and simpler code. Note: 'char'
 * is not type-compatible with 'signed char', and we define a separate case.
 */
#define __scalar_type_to_expr_cases(type)				\
		unsigned type:	(unsigned type)0,			\
		signed type:	(signed type)0

#define __unqual_scalar_typeof(x) typeof(				\
		_Generic((x),						\
			 char:	(char)0,				\
			 __scalar_type_to_expr_cases(char),		\
			 __scalar_type_to_expr_cases(short),		\
			 __scalar_type_to_expr_cases(int),		\
			 __scalar_type_to_expr_cases(long),		\
			 __scalar_type_to_expr_cases(long long),	\
			 default: (x)))

/* Is this type a native word size -- useful for atomic operations */
#define __native_word(t) \
	(sizeof(t) == sizeof(char) || sizeof(t) == sizeof(short) || \
	 sizeof(t) == sizeof(int) || sizeof(t) == sizeof(long))

/* Compile time object size, -1 for unknown */
#ifndef __compiletime_object_size
# define __compiletime_object_size(obj) -1
#endif

#ifdef __OPTIMIZE__
# define __compiletime_assert(condition, msg, prefix, suffix)		\
	do {								\
		extern void prefix ## suffix(void) __compiletime_error(msg); \
		if (!(condition))					\
			prefix ## suffix();				\
	} while (0)
#else
# define __compiletime_assert(condition, msg, prefix, suffix) do { } while (0)
#endif

#define _compiletime_assert(condition, msg, prefix, suffix) \
	__compiletime_assert(condition, msg, prefix, suffix)

/**
 * compiletime_assert - break build and emit msg if condition is false
 * @condition: a compile-time constant condition to check
 * @msg:       a message to emit if condition is false
 *
 * In tradition of POSIX assert, this macro will break the build if the
 * supplied condition is *false*, emitting the supplied error message if the
 * compiler has support to do so.
 */
#define compiletime_assert(condition, msg) \
	_compiletime_assert(condition, msg, __compiletime_assert_, __COUNTER__)

#define compiletime_assert_atomic_type(t)				\
	compiletime_assert(__native_word(t),				\
		"Need native word sized stores/loads for atomicity.")

/* Helpers for emitting diagnostics in pragmas. */
#ifndef __diag
#define __diag(string)
#endif

#ifndef __diag_GCC
#define __diag_GCC(version, severity, string)
#endif

#define __diag_push()	__diag(push)
#define __diag_pop()	__diag(pop)

#define __diag_ignore(compiler, version, option, comment) \
	__diag_ ## compiler(version, ignore, option)
#define __diag_warn(compiler, version, option, comment) \
	__diag_ ## compiler(version, warn, option)
#define __diag_error(compiler, version, option, comment) \
	__diag_ ## compiler(version, error, option)

#endif /* __LINUX_COMPILER_TYPES_H */
