#ifndef AUTOSLAB_H_INCLUDED
#define AUTOSLAB_H_INCLUDED

#include "gcc-common.h"
#include "gcc-ksyms.h"
#include <regex.h>

// usercopy attrs
extern struct attribute_spec alloc_flags_attr;
extern struct attribute_spec alloc_node_attr;
extern struct attribute_spec alloc_attr;

void autoslab_get_alloc_attrs(tree fndecl, tree *count_arg, tree *size_arg, tree *flags_arg, tree *flags_kind, tree *node_arg);

// plugin <-> target interaction
void autoslab_find_decls(void *event_data, void *data __unused);

extern GTY(()) tree autoslab_info_type;
extern GTY(()) tree autoslab_infos_type;
extern GTY(()) tree kfree_decl;
extern GTY(()) tree kzfree_decl;
extern GTY(()) tree kfree_sensitive_decl;
extern GTY(()) tree kmem_cache_create_usercopy_typename_decl;

enum kmem_call_names {
	KMEM_CACHE_ALLOC = 0,
	KMEM_CACHE_ALLOC_NODE,
	KMEM_CACHE_ALLOC_TYPE,
	KMEM_CACHE_ALLOC_NODE_TYPE,
	KMEM_CACHE_ALLOC_INDEX,
	KMEM_CACHE_ALLOC_NODE_INDEX,
	KMEM_CACHE_ALLOC_INDEX_TYPE,
	KMEM_CACHE_ALLOC_NODE_INDEX_TYPE,
	KMEM_CACHE_NUM
};

#ifdef __cplusplus
static inline enum kmem_call_names& operator++(enum kmem_call_names& kmem_call)
{
	if (kmem_call >= KMEM_CACHE_NUM)
		gcc_unreachable();

	kmem_call = static_cast<enum kmem_call_names>(static_cast<unsigned int>(kmem_call) + 1);
	return kmem_call;
}
#endif

extern GTY(()) struct kmem_cache_decl {
	const char * const name;
	tree decl;
} kmem_cache_decls[KMEM_CACHE_NUM];

extern const struct ggc_root_tab gt_ggc_r_gt_autoslab[];

extern const struct kalloc_function {
	const char *kalloc_name;
	enum kmem_call_names kmem_name;
	enum kmem_call_names kmem_name_type;
	enum kmem_call_names kmem_name_index;
	enum kmem_call_names kmem_name_index_type;
} kalloc_functions[4];

// autostack
extern bool autostack;

#define CONVERT_AUTOSTACK "convert_autostack"
extern GTY(()) tree convert_autostack_decl;

#define FREE_AUTOSTACK "free_autostack"
extern GTY(()) tree free_autostack_decl;

extern const struct ggc_root_tab gt_ggc_r_gt_autostack[];

bool can_convert_to_automatic(gimple_stmt_iterator *gsi,
			      tree size,
			      sbitmap gfp_zones,
			      bool atomic,
			      unsigned HOST_WIDE_INT lhs_align_val);
void autoslab_convert_to_automatic(gimple_stmt_iterator *gsi,
				   tree size,
				   tree flags,
				   sbitmap gfp_zones,
				   unsigned HOST_WIDE_INT lhs_size_val,
				   unsigned HOST_WIDE_INT lhs_align_val);

// autotypename
extern regex_t sizeof_regex;

// misc
extern bool verbose;
tree autoslab_build_var(location_t loc, tree id, tree type, tree ctor);
bool is_cache_alloc(const char *fnname);
tree autoslab_get_alloc_type(gcall *kalloc_call_stmt);

// gcc passes
#if BUILDING_GCC_VERSION >= 4009
gimple_opt_pass *make_pass_autoslab(gcc::context *);
simple_ipa_opt_pass *make_pass_autotypename(gcc::context *);
#else
struct opt_pass *make_pass_autoslab(void);
struct opt_pass *make_pass_autotypename(void);
#endif

#endif
