/*
 * Copyright 2019-2021 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 */

#include "autoslab.h"

GTY(()) tree autoslab_info_type;
GTY(()) tree autoslab_infos_type;
GTY(()) tree kfree_decl;
GTY(()) tree kzfree_decl;
GTY(()) tree kfree_sensitive_decl;
GTY(()) tree kmem_cache_create_usercopy_typename_decl;

GTY(()) struct kmem_cache_decl kmem_cache_decls[KMEM_CACHE_NUM] = {
#define INIT_KMEM_CACHE_DECLS(NAME, _name)	\
[KMEM_CACHE_##NAME] = {				\
		.name = "kmem_cache_" _name,	\
		.decl = NULL_TREE,		\
}

	INIT_KMEM_CACHE_DECLS(ALLOC,                 "alloc"),
	INIT_KMEM_CACHE_DECLS(ALLOC_NODE,            "alloc_node"),
	INIT_KMEM_CACHE_DECLS(ALLOC_TYPE,            "alloc_type"),
	INIT_KMEM_CACHE_DECLS(ALLOC_NODE_TYPE,       "alloc_node_type"),
	INIT_KMEM_CACHE_DECLS(ALLOC_INDEX,           "alloc_index"),
	INIT_KMEM_CACHE_DECLS(ALLOC_NODE_INDEX,      "alloc_node_index"),
	INIT_KMEM_CACHE_DECLS(ALLOC_INDEX_TYPE,      "alloc_index_type"),
	INIT_KMEM_CACHE_DECLS(ALLOC_NODE_INDEX_TYPE, "alloc_node_index_type"),

#undef INIT_KMEM_CACHE_DECLS
};

EXPORTED_CONST struct ggc_root_tab gt_ggc_r_gt_autoslab[] = {
	{
		.base = &kmem_cache_decls[0].decl,
		.nelt = ARRAY_SIZE(kmem_cache_decls),
		.stride = sizeof(kmem_cache_decls[0]),
		.cb = &gt_ggc_mx_tree_node,
		.pchw = &gt_pch_nx_tree_node
	},
	{
		.base = &autoslab_info_type,
		.nelt = 1,
		.stride = sizeof(autoslab_info_type),
		.cb = &gt_ggc_mx_tree_node,
		.pchw = &gt_pch_nx_tree_node
	},
	{
		.base = &autoslab_infos_type,
		.nelt = 1,
		.stride = sizeof(autoslab_infos_type),
		.cb = &gt_ggc_mx_tree_node,
		.pchw = &gt_pch_nx_tree_node
	},
	{
		.base = &kfree_decl,
		.nelt = 1,
		.stride = sizeof(kfree_decl),
		.cb = &gt_ggc_mx_tree_node,
		.pchw = &gt_pch_nx_tree_node
	},
	{
		.base = &kzfree_decl,
		.nelt = 1,
		.stride = sizeof(kzfree_decl),
		.cb = &gt_ggc_mx_tree_node,
		.pchw = &gt_pch_nx_tree_node
	},
	{
		.base = &kfree_sensitive_decl,
		.nelt = 1,
		.stride = sizeof(kfree_sensitive_decl),
		.cb = &gt_ggc_mx_tree_node,
		.pchw = &gt_pch_nx_tree_node
	},
	{
		.base = &kmem_cache_create_usercopy_typename_decl,
		.nelt = 1,
		.stride = sizeof(kmem_cache_create_usercopy_typename_decl),
		.cb = &gt_ggc_mx_tree_node,
		.pchw = &gt_pch_nx_tree_node
	},
	LAST_GGC_ROOT_TAB
};

const struct kalloc_function kalloc_functions[4] = {
	{
		.kalloc_name          = "kmalloc",
		.kmem_name            = KMEM_CACHE_ALLOC,
		.kmem_name_type       = KMEM_CACHE_ALLOC_TYPE,
		.kmem_name_index      = KMEM_CACHE_ALLOC_INDEX,
		.kmem_name_index_type = KMEM_CACHE_ALLOC_INDEX_TYPE,
	},
	{
		.kalloc_name          = "kmalloc_node",
		.kmem_name            = KMEM_CACHE_ALLOC_NODE,
		.kmem_name_type       = KMEM_CACHE_ALLOC_NODE_TYPE,
		.kmem_name_index      = KMEM_CACHE_ALLOC_NODE_INDEX,
		.kmem_name_index_type = KMEM_CACHE_ALLOC_NODE_INDEX_TYPE,
	},
	{
		.kalloc_name          = "__kmalloc_track_caller",
		.kmem_name            = KMEM_CACHE_ALLOC,
		.kmem_name_type       = KMEM_CACHE_ALLOC_TYPE,
		.kmem_name_index      = KMEM_CACHE_ALLOC_INDEX,
		.kmem_name_index_type = KMEM_CACHE_ALLOC_INDEX_TYPE,
	},
	{
		.kalloc_name          = "__kmalloc_node_track_caller",
		.kmem_name            = KMEM_CACHE_ALLOC_NODE,
		.kmem_name_type       = KMEM_CACHE_ALLOC_NODE_TYPE,
		.kmem_name_index      = KMEM_CACHE_ALLOC_NODE_INDEX,
		.kmem_name_index_type = KMEM_CACHE_ALLOC_NODE_INDEX_TYPE,
	},
};

static void __autoslab_analyze_size(tree size, unsigned int *c1_out, unsigned int *c0_out);

static size_t is_kalloc(const char *fnname)
{
	size_t i;

	for (i = 0; i < ARRAY_SIZE(kalloc_functions); i++)
		if (!strcmp(fnname, kalloc_functions[i].kalloc_name))
			break;
	return i;
}

bool is_cache_alloc(const char *fnname)
{
	size_t i;

	for (i = 0; i < ARRAY_SIZE(kmem_cache_decls); i++)
		if (!strcmp(fnname, kmem_cache_decls[i].name))
			return true;
	return false;
}

tree autoslab_build_var(location_t loc, tree id, tree type, tree ctor)
{
	tree var;

	var = build_decl(loc, VAR_DECL, id, type);
	DECL_INITIAL(var) = ctor;
	DECL_EXTERNAL(var) = 0;
	DECL_IGNORED_P(var) = 0;
	DECL_ARTIFICIAL(var) = 1;
	SET_DECL_ALIGN(var, UNITS_PER_WORD * BITS_PER_UNIT);
	DECL_USER_ALIGN(var) = 1;

	TREE_USED(var) = 1;

	add_referenced_var(var);

	return var;
}

static tree autoslab_build_global_var(location_t loc, tree id, tree type, tree ctor, const char *section)
{
	tree var;

	var = autoslab_build_var(loc, id, type, ctor);

	DECL_CONTEXT(var) = DECL_CONTEXT(current_function_decl);
	DECL_PRESERVE_P(var) = 1;

	TREE_ADDRESSABLE(var) = 1;
	TREE_STATIC(var) = 1;
	TREE_PUBLIC(var) = 1;

	if (section)
		set_decl_section_name(var, section);
	else
		TREE_READONLY(var) = 1;

	varpool_add_new_variable(var);
#if BUILDING_GCC_VERSION <= 4008
	if (!flag_toplevel_reorder)
		varpool_assemble_decl(varpool_node_for_decl(var));
#endif

	return var;
}

static tree autoslab_build_cache_ptr(location_t loc, tree id, tree type, tree ctor)
{
	const char *section_name;
	tree section;

	section = lookup_attribute("section", DECL_ATTRIBUTES(current_function_decl));
	if (section && TREE_VALUE(section)) {
		section = TREE_VALUE(TREE_VALUE(section));

		if (!strncmp(TREE_STRING_POINTER(section), ".init.text", 10))
			section_name = ".data.init.autoslab";
		else if (!strncmp(TREE_STRING_POINTER(section), ".meminit.text", 13))
			section_name = ".data.init.autoslab";
		else
			section_name = ".data.autoslab";
		if (verbose)
			inform(loc, "autoslab: yes section %E %E", section, id);
	} else {
		section_name = ".data.autoslab";
		if (verbose)
			inform(loc, "autoslab: yes section \".text\" %E", id);
	}

	return autoslab_build_global_var(loc, id, type, ctor, section_name);
}

static tree __autoslab_get_alloc_type(gcall *kalloc_call_stmt, gimple_stmt_ptr use_stmt)
{
	tree lhs, lhs_type;

	switch (gimple_code(use_stmt)) {
	default:
		print_gimple_stmt(stderr, kalloc_call_stmt, 0, TDF_LINENO);
		print_gimple_stmt(stderr, use_stmt, 0, TDF_LINENO);
		gcc_unreachable();

	case GIMPLE_ASSIGN:
		lhs = gimple_assign_lhs(use_stmt);
		break;

	case GIMPLE_CALL:
		lhs = gimple_call_lhs(kalloc_call_stmt);
		break;

	case GIMPLE_PHI:
		lhs = gimple_phi_result(use_stmt);
		break;

	case GIMPLE_RETURN:
		lhs = gimple_return_retval(as_a_greturn(use_stmt));
		break;
	}

	lhs_type = TREE_TYPE(lhs);
	if (POINTER_TYPE_P(lhs_type))
		return TREE_TYPE(lhs_type);
	else
		return TREE_TYPE(gimple_call_lhs(kalloc_call_stmt));
}

tree autoslab_get_alloc_type(gcall *kalloc_call_stmt)
{
	tree lhs, lhs_type, alloc_type;
	use_operand_p use_p;
	imm_use_iterator imm_iter;
	gimple_stmt_ptr use_stmt;

	lhs = gimple_call_lhs(kalloc_call_stmt);
	gcc_assert(lhs);
	lhs_type = TREE_TYPE(lhs);
	gcc_assert(POINTER_TYPE_P(lhs_type));
	lhs_type = TREE_TYPE(lhs_type);

	if (RECORD_OR_UNION_TYPE_P(lhs_type))
		return lhs_type;

	if (single_imm_use(lhs, &use_p, &use_stmt))
		return __autoslab_get_alloc_type(kalloc_call_stmt, use_stmt);

	alloc_type = NULL_TREE;
	FOR_EACH_IMM_USE_STMT(use_stmt, imm_iter, lhs) {
		tree type_size;

		if (!is_gimple_assign(use_stmt))
			continue;
		if (gimple_assign_rhs_code(use_stmt) != SSA_NAME)
			continue;
		if (gimple_assign_rhs1(use_stmt) != lhs) {
			lhs_type = gimple_assign_lhs(use_stmt);
			lhs_type = get_inner_ref(lhs_type);
			switch (TREE_CODE(lhs_type)) {
			default:
				debug_tree(lhs);
				debug_tree(lhs_type);
				gcc_unreachable();

#if BUILDING_GCC_VERSION >= 4006
			case MEM_REF:
				gcc_assert(TREE_OPERAND(lhs_type, 0) == lhs);
				if (!integer_zerop(TREE_OPERAND(lhs_type, 1)))
					continue;
				break;
#endif

			case TARGET_MEM_REF:
				gcc_assert(TMR_BASE(lhs_type) == lhs);
				if (TMR_OFFSET(lhs_type) && !integer_zerop(TMR_OFFSET(lhs_type)))
					continue;
				break;
			}
		} else {
			lhs_type = TREE_TYPE(gimple_assign_lhs(use_stmt));
			if (!POINTER_TYPE_P(lhs_type))
				continue;
		}

		lhs_type = TREE_TYPE(lhs_type);
		if (!RECORD_OR_UNION_TYPE_P(lhs_type))
			continue;

		type_size = TYPE_SIZE(lhs_type);
		if (!type_size)
			continue;

		gcc_assert(TREE_CODE(type_size) == INTEGER_CST);
		if (!alloc_type || tree_int_cst_compare(type_size, TYPE_SIZE(alloc_type)) > 0)
			alloc_type = lhs_type;
	}

	return alloc_type ?: TREE_TYPE(TREE_TYPE(lhs));
}

static int autoslab_gen_name(char **name,
			     const char *kind,
			     const char *type,
			     gimple_stmt_iterator *gsi,
			     unsigned HOST_WIDE_INT size_val,
			     unsigned HOST_WIDE_INT lhs_align_val,
			     unsigned HOST_WIDE_INT flags_val,
			     sbitmap gfp_zones,
			     unsigned int minshift,
			     unsigned int maxshift,
			     unsigned int c1,
			     unsigned int c0)
{
	static unsigned int counter;
	gimple_stmt_ptr stmt;
	location_t loc1, loc2;
	char temp[TREE_INT_CST_LOW(KSYM_NAME_LEN)];
	int len;

	stmt = gsi_stmt(*gsi);
	loc1 = gimple_inline_point_location(stmt);
	loc2 = gimple_expansion_point_location(stmt);

	len = snprintf(temp, sizeof temp,
		"autoslab_%s%s"
		"M_%s_P_%s_%u_%u_%u_%u_S_"
		HOST_WIDE_INT_PRINT_UNSIGNED "_A_" HOST_WIDE_INT_PRINT_UNSIGNED
		"_%s",
		kind, type,
		TREE_STRING_POINTER(KBUILD_MODNAME),
		TREE_STRING_POINTER(AUTOSLAB_BASENAME),
		LOCATION_LINE(loc1),
		LOCATION_COLUMN(loc1),
		LOCATION_LINE(loc2),
		LOCATION_COLUMN(loc2),
		size_val,
		lhs_align_val,
		sbitmap_count_bits(gfp_zones) > 1 ? "i" :
		ZONE_DMA && sbitmap_bit_p(gfp_zones, TREE_INT_CST_LOW(ZONE_DMA)) ? "d" :
		ZONE_DMA32 && sbitmap_bit_p(gfp_zones, TREE_INT_CST_LOW(ZONE_DMA32)) ? "D" :
		!sbitmap_bit_p(gfp_zones, TREE_INT_CST_LOW(ZONE_NORMAL)) ? NULL :
		flags_val & TREE_INT_CST_LOW(__GFP_RECLAIMABLE) ? "r" : "n");

	if (len == -1 || (unsigned long)len >= sizeof temp)
		return -1;

	if (strcmp(kind, "sizes_"))
		len = asprintf(name, "%s_%u", temp, counter++);
	else
		len = asprintf(name, "%s_MIN_%u_MAX_%u_C1_%u_C0_%u_%u",
			temp,
			minshift,
			maxshift,
			c1,
			c0,
			counter++);

	if (len == -1)
		return -1;

	if ((unsigned HOST_WIDE_INT)len < TREE_INT_CST_LOW(KSYM_NAME_LEN))
		return len;

	if (seen_error())
		return len;

	error_at(loc2, "autoslab: increase KSYM_NAME_LEN to %qd at least because of %qs!", len, *name);
	return len;
}

static tree autoslab_create_info(gimple_stmt_iterator *gsi,
				 tree size,
				 tree flags,
				 sbitmap gfp_zones,
				 const char *type,
				 const char *lhs_type_name_ptr,
				 unsigned HOST_WIDE_INT lhs_align_val)
{
	gcall *kalloc_call_stmt;
	tree field, value, ctor, autoslab_info;
	unsigned HOST_WIDE_INT size_val, flags_val;
	char *autoslab_info_name, *autoslab_name;
	int autoslab_name_len;
	varpool_node_ptr var;
	location_t loc;
#if BUILDING_GCC_VERSION <= 4007
	VEC(constructor_elt, gc) *initializer = NULL;
#else
	vec<constructor_elt, va_gc> *initializer = NULL;
#endif

	kalloc_call_stmt = as_a_gcall(gsi_stmt(*gsi));
	loc = gimple_expansion_point_location(kalloc_call_stmt);
	size_val = TREE_INT_CST_LOW(size);
	flags_val = TREE_CODE(flags) == INTEGER_CST ? TREE_INT_CST_LOW(flags) : 0;

	// build:
	// const struct autoslab_info i __initdata = &(struct autoslab_info){"name", "typename", size, align, flags, useroffset, usersize};
	// struct autoslab_info *pi = &i;

	// construct the initializer for autoslab_info
	gcc_assert(autoslab_info_type);
	field = TYPE_FIELDS(autoslab_info_type);
#if BUILDING_GCC_VERSION <= 4007
	initializer = VEC_alloc(constructor_elt, gc, list_length(field));
#else
	vec_alloc(initializer, list_length(field));
#endif

	// .name
	autoslab_name_len = autoslab_gen_name(&autoslab_name, "", type, gsi, size_val, lhs_align_val, flags_val, gfp_zones, 0, 0, 0, 0);
	gcc_assert(autoslab_name_len != -1);
	value = build_const_char_string(autoslab_name_len + 1, autoslab_name);
	free(autoslab_name);
	value = build1_loc(loc, ADDR_EXPR, TREE_TYPE(field), value);
	CONSTRUCTOR_APPEND_ELT(initializer, field, value);

	// .typename
	field = TREE_CHAIN(field);
	value = build_const_char_string(strlen(lhs_type_name_ptr) + 1, lhs_type_name_ptr);
	value = build1_loc(loc, ADDR_EXPR, TREE_TYPE(field), value);
	CONSTRUCTOR_APPEND_ELT(initializer, field, value);

	// .size
	field = TREE_CHAIN(field);
	value = fold_binary_to_constant(MAX_EXPR, TREE_TYPE(field), size, build_int_cst(TREE_TYPE(field), 8));
	gcc_assert(value);
	CONSTRUCTOR_APPEND_ELT(initializer, field, value);

	// .align
	field = TREE_CHAIN(field);
	value = fold_binary_to_constant(MAX_EXPR, TREE_TYPE(field), build_int_cst(TREE_TYPE(field), lhs_align_val), ARCH_KMALLOC_MINALIGN);
	gcc_assert(value);
	CONSTRUCTOR_APPEND_ELT(initializer, field, value);

	// .flags
	field = TREE_CHAIN(field);
	if (flags_val & tree_to_uhwi(__GFP_DMA))
		value = fold_convert_loc(loc, TREE_TYPE(field), SLAB_CACHE_DMA);
	else if (flags_val & tree_to_uhwi(__GFP_DMA32))
		value = fold_convert_loc(loc, TREE_TYPE(field), SLAB_CACHE_DMA32);
	else if (flags_val & tree_to_uhwi(__GFP_RECLAIMABLE))
		value = fold_convert_loc(loc, TREE_TYPE(field), SLAB_RECLAIM_ACCOUNT);
	else if (flags_val & tree_to_uhwi(__GFP_USERCOPY))
		value = fold_convert_loc(loc, TREE_TYPE(field), SLAB_USERCOPY);
	else
		value = fold_convert_loc(loc, TREE_TYPE(field), integer_zero_node);
	CONSTRUCTOR_APPEND_ELT(initializer, field, value);

	// .useroffset
	field = TREE_CHAIN(field);
	value = build_int_cst(TREE_TYPE(field), 0);
	CONSTRUCTOR_APPEND_ELT(initializer, field, value);

	// .usersize
	field = TREE_CHAIN(field);
	value = fold_convert_loc(loc, TREE_TYPE(field), size);
	CONSTRUCTOR_APPEND_ELT(initializer, field, value);

	ctor = build_constructor(autoslab_info_type, initializer);

	TREE_STATIC(ctor) = 1;
	TREE_PUBLIC(ctor) = 0;
	TREE_CONSTANT(ctor) = 1;
	TREE_USED(ctor) = 1;

	gcc_assert(-1 != autoslab_gen_name(&autoslab_info_name, "info_", type, gsi, size_val, lhs_align_val, flags_val, gfp_zones, 0, 0, 0, 0));
	var = varpool_get_for_asmname(get_identifier(autoslab_info_name));
	gcc_assert(!var);
	autoslab_info = autoslab_build_global_var(loc, get_identifier(autoslab_info_name), autoslab_info_type, ctor, ".init.data.autoslab");
	free(autoslab_info_name);

	return autoslab_info;
}

#if BUILDING_GCC_VERSION <= 4007
typedef VEC(constructor_elt, gc) *ctor_values;
#else
typedef vec<constructor_elt, va_gc> *ctor_values;
#endif

static ctor_values autoslab_create_infos_initializer(gimple_stmt_iterator *gsi,
						     tree size,
						     sbitmap gfp_zones,
				 		     const char *lhs_type_name_ptr,
						     unsigned HOST_WIDE_INT lhs_align_val)
{
	gcall *kalloc_call_stmt;
	tree field, value;
	location_t loc;
#if BUILDING_GCC_VERSION <= 4007
	ctor_values ctor_values = VEC_alloc(constructor_elt, gc, list_length(TYPE_FIELDS(autoslab_infos_type)));
#else
	ctor_values ctor_values = NULL;
	vec_alloc(ctor_values, list_length(TYPE_FIELDS(autoslab_infos_type)));
#endif

	kalloc_call_stmt = as_a_gcall(gsi_stmt(*gsi));
	loc = gimple_expansion_point_location(kalloc_call_stmt);

	gcc_assert(autoslab_infos_type);
	for (field = TYPE_FIELDS(autoslab_infos_type); field != NULL_TREE; field = TREE_CHAIN(field)) {
		tree flags;
		const char *type;

		if (id_equal(DECL_NAME(field), "normal")) {
			// .normal
			flags = integer_zero_node;
			type = "normal_";
#if 0
		} else if (id_equal(DECL_NAME(field), "reclaimable")) {
			// .reclaimable
			flags = __GFP_RECLAIMABLE;
			type = "reclaimable_";
#endif
		} else if (id_equal(DECL_NAME(field), "dma")) {
			// .dma
			flags = __GFP_DMA;
			type = "dma_";
		} else if (id_equal(DECL_NAME(field), "dma32")) {
			// .dma32
			flags = __GFP_DMA32;
			type = "dma32_";
		} else {
			debug_tree(autoslab_infos_type);
			debug_tree(field);
			gcc_unreachable();
		}

		value = autoslab_create_info(gsi, size, flags, gfp_zones, type, lhs_type_name_ptr, lhs_align_val);
		value = build1_loc(loc, ADDR_EXPR, TYPE_POINTER_TO(autoslab_info_type), value);
		CONSTRUCTOR_APPEND_ELT(ctor_values, field, value);
	}

	return ctor_values;
}

static tree autoslab_create_infos_type(gimple_stmt_iterator *gsi,
				       tree size,
				       sbitmap gfp_zones,
				       const char *lhs_type_name_ptr,
				       unsigned HOST_WIDE_INT lhs_align_val)
{
	gcall *kalloc_call_stmt;
	tree ctor, autoslab_infos;
	unsigned HOST_WIDE_INT size_val;
	char *autoslab_infos_name;
	varpool_node_ptr var;
	location_t loc;

	kalloc_call_stmt = as_a_gcall(gsi_stmt(*gsi));
	loc = gimple_expansion_point_location(kalloc_call_stmt);
	size_val = TREE_INT_CST_LOW(size);

	gcc_assert(-1 != autoslab_gen_name(&autoslab_infos_name, "infos_", "", gsi, size_val, lhs_align_val, 0, gfp_zones, 0, 0, 0, 0));
	var = varpool_get_for_asmname(get_identifier(autoslab_infos_name));
	gcc_assert(!var);

	// build:
	// struct autoslab_infos autoslabs = { .normal = autoslab_normal, .dma = autoslab_dma, .dma32 = autoslab_dma32 };

	ctor = build_constructor(autoslab_infos_type, autoslab_create_infos_initializer(gsi, size, gfp_zones, lhs_type_name_ptr, lhs_align_val));

	autoslab_infos = autoslab_build_cache_ptr(loc, get_identifier(autoslab_infos_name), autoslab_infos_type, ctor);
	free(autoslab_infos_name);

	return autoslab_infos;
}

static tree autoslab_analyze_size(gimple_stmt_iterator *gsi,
				  tree size,
				  unsigned int *c1,
				  unsigned int *c0,
				  unsigned int *minshift,
				  unsigned int *maxshift)
{
	gcall *kalloc_call_stmt;
	unsigned HOST_WIDE_INT kmalloc_max, minsize, maxsize, maxsize_ceil;

	gcc_assert(TREE_CODE(size) == INTEGER_CST || TREE_CODE(size) == SSA_NAME);

	kalloc_call_stmt = as_a_gcall(gsi_stmt(*gsi));

	__autoslab_analyze_size(size, c1, c0);

	if (*c1 == 0) {
		*minshift = ceil_log2(*c0);
		*maxshift = ceil_log2(*c0);

		if (verbose)
			inform(gimple_inline_point_location(kalloc_call_stmt),
			       "autoslab: CONSTSIZE"
			       " minshift %u maxshift %u c1 %u c0 %u",
			       *minshift, *maxshift, *c1, *c0);

		return build_int_cstu(sizetype, *c0);
	}

	minsize = TREE_INT_CST_LOW(KMALLOC_MIN_SIZE);
	maxsize = kmalloc_max = TREE_INT_CST_LOW(KMALLOC_MAX_SIZE);

#if BUILDING_GCC_VERSION >= 4009
	wide_int minrange, maxrange;
	location_t loc = gimple_expansion_point_location(kalloc_call_stmt);

	switch (get_range_info(size, &minrange, &maxrange)) {
	default:
		break;

	case VR_ANTI_RANGE:
		if (minrange.to_uhwi() >= kmalloc_max + 1)
			break;
		if (maxrange.to_uhwi() <= kmalloc_max - 1)
			break;

		maxsize = minrange.to_uhwi() - 1;
		if (minsize > maxsize)
			minsize = maxsize;
		if (verbose)
			inform(loc, "autoslab: SIZE ANTIRANGE 0-%lu,%lu-inf c1 %u c0 %u",
				minrange.to_uhwi() - 1, maxrange.to_uhwi() + 1, *c1, *c0);
		break;

	case VR_RANGE:
		if (maxrange.to_uhwi() >= kmalloc_max)
			break;

		if (minrange.to_uhwi() > minsize)
			minsize = minrange.to_uhwi();
		if (maxrange.to_uhwi() > minsize)
			maxsize = maxrange.to_uhwi();
		else
			maxsize = minsize;

		if (verbose)
			inform(loc, "autoslab: SIZE RANGE %lu-%lu c1 %u c0 %u",
				minrange.to_uhwi(), maxrange.to_uhwi(), *c1, *c0);
		break;
	}
#endif

	if (*c0 > maxsize / 2 || *c0 + *c1 > maxsize / 2) {
		*c0 = 0;
		*c1 = 1;
	}

	*minshift = minsize < *c0 + *c1 ? ceil_log2(*c1 + *c0) : ceil_log2(minsize);
	gcc_assert(maxsize >= *c0 + *c1);
	*maxshift = ceil_log2(maxsize);

	if (verbose)
		inform(gimple_inline_point_location(kalloc_call_stmt),
		       "autoslab: VARSIZE minsize " HOST_WIDE_INT_PRINT_UNSIGNED
		       " maxsize " HOST_WIDE_INT_PRINT_UNSIGNED
		       " minshift %u maxshift %u c1 %u c0 %u",
		       minsize, maxsize, *minshift, *maxshift, *c1, *c0);

	maxsize_ceil = ((1UL << *maxshift) - *c0) / *c1 * *c1 + *c0;
	if (maxsize_ceil < maxsize)
		maxsize = maxsize_ceil;
	size = build_int_cstu(unsigned_type_node, maxsize);
	gcc_assert(tree_int_cst_compare(size, KMALLOC_MAX_SIZE) <= 0);
	return size;
}

static tree autoslab_create_infos_size(gimple_stmt_iterator *gsi,
				       tree size,
				       tree flags,
				       sbitmap gfp_zones,
				       const char *lhs_type_name_ptr,
				       unsigned int c1,
				       unsigned int c0,
				       unsigned int minshift,
				       unsigned int maxshift,
				       unsigned HOST_WIDE_INT maxsize,
				       unsigned HOST_WIDE_INT lhs_align_val)
{
	tree ctor, autoslab_infos;
	char *autoslab_infos_name;
	varpool_node_ptr var;
	location_t loc;
	unsigned HOST_WIDE_INT flags_val;
	unsigned int i;

	loc = gimple_expansion_point_location(gsi_stmt(*gsi));
	flags_val = TREE_CODE(flags) == INTEGER_CST ? TREE_INT_CST_LOW(flags) : 0;
	gcc_assert(-1 != autoslab_gen_name(&autoslab_infos_name, "sizes_", "", gsi, 0, lhs_align_val, flags_val, gfp_zones, minshift, maxshift, c1, c0));
	var = varpool_get_for_asmname(get_identifier(autoslab_infos_name));
	gcc_assert(!var);

#if BUILDING_GCC_VERSION <= 4007
	ctor_values initializer = VEC_alloc(constructor_elt, gc, maxshift - minshift + 1);
#else
	ctor_values initializer = NULL;
	vec_alloc(initializer, maxshift - minshift + 1);
#endif

	// build:
	// struct autoslab_infos *autoslabs = { [i] = &autoslab_size_<2^i/c1*c1>, ... };

	// loop over all sizes and create the autoslab_info for each
	for (i = minshift; i <= maxshift; i++) {
		unsigned int size;

		size = ((1UL << i) - c0) / c1 * c1 + c0;
		if (size > maxsize)
			size = maxsize;
		gcc_assert(size <= TREE_INT_CST_LOW(KMALLOC_MAX_SIZE));

		if (sbitmap_count_bits(gfp_zones) == 1) {
			tree value;

			value = autoslab_create_info(gsi, build_int_cstu(long_unsigned_type_node, size), flags, gfp_zones, "size_", lhs_type_name_ptr, lhs_align_val);
			value = build1_loc(loc, ADDR_EXPR, TYPE_POINTER_TO(autoslab_info_type), value);
			CONSTRUCTOR_APPEND_ELT(initializer, build_int_cstu(unsigned_type_node, i - minshift), value);
		} else {
			ctor_values values;
			tree value;
			unsigned int j;

			values = autoslab_create_infos_initializer(gsi, build_int_cstu(long_unsigned_type_node, size), gfp_zones, lhs_type_name_ptr, lhs_align_val);
			FOR_EACH_CONSTRUCTOR_VALUE(values, j, value)
#if BUILDING_GCC_VERSION <= 4007
				CONSTRUCTOR_APPEND_ELT(initializer, build_int_cstu(unsigned_type_node, VEC_length(constructor_elt, values) * (i - minshift) + j), value);
#else
				CONSTRUCTOR_APPEND_ELT(initializer, build_int_cstu(unsigned_type_node, values->length() * (i - minshift) + j), value);
#endif
		}
	}

#if BUILDING_GCC_VERSION <= 4007
	ctor = build_constructor(build_array_type_nelts(TYPE_POINTER_TO(autoslab_info_type), VEC_length(constructor_elt, initializer)), initializer);
#else
	ctor = build_constructor(build_array_type_nelts(TYPE_POINTER_TO(autoslab_info_type), initializer->length()), initializer);
#endif
	autoslab_infos = autoslab_build_cache_ptr(loc, get_identifier(autoslab_infos_name), TREE_TYPE(ctor), ctor);

	free(autoslab_infos_name);

	return autoslab_infos;
}

static void __autoslab_convert_call(gimple_stmt_iterator *gsi,
				    enum kmem_call_names kmem_call,
				    tree size,
				    tree flags,
				    sbitmap gfp_zones,
				    tree node,
				    tree minshift,
				    tree maxshift,
				    const char *lhs_type_name_ptr,
				    unsigned HOST_WIDE_INT lhs_size_val,
				    unsigned HOST_WIDE_INT lhs_align_val,
				    tree kmem_cache)
{
	gcall *kalloc_call_stmt, *kmem_call_stmt;
	tree autoslab, kmem_cache_alloc;
	unsigned HOST_WIDE_INT size_val;
	location_t loc1, loc2;
	const char *autoslab_kind;
	cgraph_node_ptr callee;
	basic_block bb;
	int frequency;

	kalloc_call_stmt = as_a_gcall(gsi_stmt(*gsi));
	loc1 = gimple_inline_point_location(kalloc_call_stmt);
	loc2 = gimple_expansion_point_location(kalloc_call_stmt);
	size_val = TREE_CONSTANT(size) ? TREE_INT_CST_LOW(size) : 0;

	// handle the allocation of a single object
	if (DECL_P(kmem_cache))
		autoslab = create_tmp_var(TREE_TYPE(kmem_cache), DECL_NAME_POINTER(kmem_cache));
	else
		autoslab = create_tmp_var(TREE_TYPE(kmem_cache), DECL_NAME_POINTER(TREE_OPERAND(TREE_OPERAND(kmem_cache, 0), 0)));
	add_referenced_var(autoslab);
	autoslab = make_ssa_name(autoslab, NULL);

	gimple_stmt_ptr assign = gimple_build_assign(autoslab, kmem_cache);
	gimple_set_location(assign, loc2);
	gimple_set_block(assign, gimple_block(kalloc_call_stmt));
	gsi_insert_before(gsi, assign, GSI_SAME_STMT);
	update_stmt(assign);

	kmem_cache_alloc = kmem_cache_decls[kmem_call].decl;
	switch (kmem_call) {
	default:
		gcc_unreachable();

	case KMEM_CACHE_ALLOC:
	case KMEM_CACHE_ALLOC_TYPE:
		kmem_call_stmt = as_a_gcall(gimple_build_call(kmem_cache_alloc, 2, autoslab, flags));
		break;

	case KMEM_CACHE_ALLOC_INDEX:
	case KMEM_CACHE_ALLOC_INDEX_TYPE:
		kmem_call_stmt = as_a_gcall(gimple_build_call(kmem_cache_alloc, 5, autoslab, size, flags, minshift, maxshift));
		break;

	case KMEM_CACHE_ALLOC_NODE:
	case KMEM_CACHE_ALLOC_NODE_TYPE:
		kmem_call_stmt = as_a_gcall(gimple_build_call(kmem_cache_alloc, 3, autoslab, flags, node));
		break;

	case KMEM_CACHE_ALLOC_NODE_INDEX:
	case KMEM_CACHE_ALLOC_NODE_INDEX_TYPE:
		kmem_call_stmt = as_a_gcall(gimple_build_call(kmem_cache_alloc, 6, autoslab, size, flags, node, minshift, maxshift));
		break;
	}

	gimple_call_copy_flags(kmem_call_stmt, kalloc_call_stmt);
	gimple_set_location(kmem_call_stmt, loc2);
	gimple_set_block(kmem_call_stmt, gimple_block(kalloc_call_stmt));
	gimple_call_set_lhs(kmem_call_stmt, gimple_call_lhs(kalloc_call_stmt));
	gsi_replace(gsi, kmem_call_stmt, true);
	update_stmt(kmem_call_stmt);

	// update the cgraph
	bb = gsi_bb(*gsi);
	callee = cgraph_get_create_node(gimple_call_fndecl(kmem_call_stmt));
	gcc_assert(callee);
	frequency = compute_call_stmt_bb_frequency(current_function_decl, bb);
	cgraph_create_edge(cgraph_get_node(current_function_decl), callee, kmem_call_stmt, bb->count, frequency, bb->loop_depth);

	if (!verbose)
		return;

	if (sbitmap_count_bits(gfp_zones) == 1)
		autoslab_kind = TREE_CODE(size) == INTEGER_CST ? "constsize" : "varsize";
	else
		autoslab_kind = TREE_CODE(size) == INTEGER_CST ? "varflags" : "varsizeflags";

	inform(loc2, "autoslab: yes %s "
		HOST_WIDE_INT_PRINT_UNSIGNED "/" HOST_WIDE_INT_PRINT_HEX_PURE " "
		HOST_WIDE_INT_PRINT_UNSIGNED "/" HOST_WIDE_INT_PRINT_HEX_PURE " "
		HOST_WIDE_INT_PRINT_UNSIGNED "/" HOST_WIDE_INT_PRINT_HEX_PURE " "
		HOST_WIDE_INT_PRINT_UNSIGNED "/" HOST_WIDE_INT_PRINT_HEX_PURE " "
		HOST_WIDE_INT_PRINT_HEX_PURE " "
		"%lx/%lx "
		HOST_WIDE_INT_PRINT_UNSIGNED "/" HOST_WIDE_INT_PRINT_HEX_PURE " "
		"%qs %u:%u %u:%u",
		autoslab_kind,
		lhs_size_val ? size_val/lhs_size_val : 0, lhs_size_val ? size_val/lhs_size_val : 0,
		size_val, size_val,
		lhs_size_val, lhs_size_val,
		lhs_align_val, lhs_align_val,
		TREE_CODE(flags) == INTEGER_CST ? TREE_INT_CST_LOW(flags) : ~0UL,
		0UL, 0UL, size_val, size_val, lhs_type_name_ptr,
		LOCATION_LINE(loc1), LOCATION_COLUMN(loc1),
		LOCATION_LINE(loc2), LOCATION_COLUMN(loc2));
}

static void autoslab_analyze_plus(tree rhs1, tree rhs2, unsigned int *c1_out, unsigned int *c0_out)
{
	unsigned int c1_in_1, c1_in_2, c0_in_1, c0_in_2;

	__autoslab_analyze_size(rhs1, &c1_in_1, &c0_in_1);
	__autoslab_analyze_size(rhs2, &c1_in_2, &c0_in_2);

	*c0_out = c0_in_1 + c0_in_2;
	if (!c1_in_1 != !c1_in_2)
		*c1_out = c1_in_1 + c1_in_2;
}

static void autoslab_analyze_mult(tree rhs1, tree rhs2, unsigned int *c1_out, unsigned int *c0_out)
{
	unsigned int c1_in_1, c1_in_2, c0_in_1, c0_in_2;

	__autoslab_analyze_size(rhs1, &c1_in_1, &c0_in_1);
	__autoslab_analyze_size(rhs2, &c1_in_2, &c0_in_2);

	gcc_assert(!(c1_in_1 > 1 && c1_in_2 > 1));

	if (c0_in_1 || c0_in_2) {
		*c0_out = c0_in_1 * c0_in_2;
		*c1_out = c0_in_1 * c1_in_2 + c0_in_2 * c1_in_1;
	} else if (c1_in_1 && c1_in_2)
		*c1_out = c1_in_1 * c1_in_2;
}

static void __autoslab_analyze_size(tree size, unsigned int *c1_out, unsigned int *c0_out)
{
	switch (TREE_CODE(size)) {
	default:
		debug_tree(size);
		gcc_unreachable();

	case ADDR_EXPR:
		*c0_out = 0;
		*c1_out = 0;
		break;

	case INTEGER_CST: {
		unsigned int c0_in;

		c0_in = TREE_INT_CST_LOW(size);
		*c0_out = c0_in <= INT_MAX ? c0_in : 0;
		*c1_out = 0;
		break;
	}

	case SSA_NAME: {
		gimple_stmt_ptr def_stmt;
		def_stmt = SSA_NAME_DEF_STMT(size);

		*c0_out = 0;
		*c1_out = 1;

		switch (gimple_code(def_stmt)) {
		default:
			inform(gimple_inline_point_location(def_stmt), "%u %u", *c1_out, *c0_out);
			debug_gimple_stmt(def_stmt);
			debug_tree(size);
			gcc_unreachable();

		case GIMPLE_ASM:
			break;

		case GIMPLE_ASSIGN:
			switch (gimple_assign_rhs_code(def_stmt)) {
			default:
				inform(gimple_inline_point_location(def_stmt), "code %s", get_tree_code_name(gimple_assign_rhs_code(def_stmt)));
				inform(gimple_expansion_point_location(def_stmt), "code %s", get_tree_code_name(gimple_assign_rhs_code(def_stmt)));
				debug_gimple_stmt(def_stmt);
				debug_tree(size);
				gcc_unreachable();

			case ABS_EXPR:
			case ADDR_EXPR:
			case ARRAY_REF:
			case BIT_FIELD_REF:
			case BIT_NOT_EXPR:
			case COND_EXPR:
			case EQ_EXPR:
			case GE_EXPR:
			case GT_EXPR:
			case INDIRECT_REF:
			case LE_EXPR:
			case LT_EXPR:
#if BUILDING_GCC_VERSION >= 4006
			case MEM_REF:
#endif
			case MIN_EXPR:
			case MAX_EXPR:
			case NE_EXPR:
			case NEGATE_EXPR:
			case PARM_DECL:
#if BUILDING_GCC_VERSION >= 8000
			case POINTER_DIFF_EXPR:
#endif
			case RROTATE_EXPR:
			case TARGET_MEM_REF:
			case VAR_DECL:
				break;

			case COMPONENT_REF: {
				tree rhs, var, field, index, value;
				unsigned int i;

				rhs = gimple_assign_rhs1(def_stmt);
				var = TREE_OPERAND(rhs, 0);
				if (!VAR_P(var) || !DECL_INITIAL(var))
					break;

				field = TREE_OPERAND(rhs, 1);
				gcc_assert(TREE_CODE(field) == FIELD_DECL);
				if (!TREE_READONLY(var) && !TREE_READONLY(field))
					break;

				FOR_EACH_CONSTRUCTOR_ELT(CONSTRUCTOR_ELTS(DECL_INITIAL(var)), i, index, value) {
					if (field == index)
						return __autoslab_analyze_size(value, c1_out, c0_out);
				}
				break;
			}

			case BIT_AND_EXPR: {
				tree rhs1, rhs2;
				unsigned HOST_WIDE_INT mask;
				unsigned int c1_in, c0_in;

				rhs1 = gimple_assign_rhs1(def_stmt);
				rhs2 = gimple_assign_rhs2(def_stmt);

				if (TREE_CODE(rhs2) != INTEGER_CST)
					break;
				mask = TREE_INT_CST_LOW(rhs2);
				if (!pow2p_hwi(mask + 1))
					break;

				__autoslab_analyze_size(rhs1, &c1_in, &c0_in);

				*c0_out = c0_in & mask;
				break;
			}

			case BIT_IOR_EXPR: {
				tree rhs1, rhs2;
				unsigned HOST_WIDE_INT mask;
				unsigned int c1_in, c0_in;

				rhs1 = gimple_assign_rhs1(def_stmt);
				rhs2 = gimple_assign_rhs2(def_stmt);

				if (TREE_CODE(rhs2) != INTEGER_CST)
					break;
				mask = TREE_INT_CST_LOW(rhs2);
				if (!pow2p_hwi(mask + 1))
					break;

				__autoslab_analyze_size(rhs1, &c1_in, &c0_in);

				*c0_out = c0_in | mask;

				break;
			}

			case LSHIFT_EXPR: {
				tree rhs1, rhs2;
				unsigned HOST_WIDE_INT lshift;
				unsigned int c1_in, c0_in;

				rhs1 = gimple_assign_rhs1(def_stmt);
				rhs2 = gimple_assign_rhs2(def_stmt);

				if (TREE_CODE(rhs2) != INTEGER_CST)
					break;

				__autoslab_analyze_size(rhs1, &c1_in, &c0_in);

				lshift = TREE_INT_CST_LOW(rhs2);
				*c0_out = c0_in << lshift;
				*c1_out = c1_in << lshift;

				break;
			}

			case MINUS_EXPR: {
				tree rhs1, rhs2;
				unsigned int c1_in, c0_in_1, c0_in_2;

				rhs1 = gimple_assign_rhs1(def_stmt);
				rhs2 = gimple_assign_rhs2(def_stmt);

				__autoslab_analyze_size(rhs1, &c1_in, &c0_in_1);

				__autoslab_analyze_size(rhs2, &c1_in, &c0_in_2);

				if (c0_in_1)
					*c0_out = c0_in_1 - c0_in_2;

				break;
			}

			case MULT_EXPR:
				autoslab_analyze_mult(gimple_assign_rhs1(def_stmt), gimple_assign_rhs2(def_stmt), c1_out, c0_out);
				break;

			case PLUS_EXPR:
			case POINTER_PLUS_EXPR:
				autoslab_analyze_plus(gimple_assign_rhs1(def_stmt), gimple_assign_rhs2(def_stmt), c1_out, c0_out);
				break;

			case EXACT_DIV_EXPR:
			case TRUNC_DIV_EXPR:
			case TRUNC_MOD_EXPR:
			case RSHIFT_EXPR:
				break;

			case REALPART_EXPR:
				__autoslab_analyze_size(TREE_OPERAND(gimple_assign_rhs1(def_stmt), 0), c1_out, c0_out);
				break;

			case COMPLEX_EXPR:
			case CONVERT_EXPR:
			case INTEGER_CST:
			case NOP_EXPR:
			case SSA_NAME:
				__autoslab_analyze_size(gimple_assign_rhs1(def_stmt), c1_out, c0_out);
				break;

			case VIEW_CONVERT_EXPR:
				__autoslab_analyze_size(TREE_OPERAND(gimple_assign_rhs1(def_stmt), 0), c1_out, c0_out);
				break;
			}
			break;

		case GIMPLE_CALL: {
			if (!gimple_call_internal_p(def_stmt))
				break;

			switch (gimple_call_internal_fn(def_stmt)) {
			default:
				debug_gimple_stmt(def_stmt);
				debug_tree(size);
				gcc_unreachable();

#if BUILDING_GCC_VERSION >= 4009
			case IFN_UBSAN_CHECK_ADD:
				autoslab_analyze_plus(gimple_call_arg(def_stmt, 0), gimple_call_arg(def_stmt, 1), c1_out, c0_out);
				break;

			case IFN_UBSAN_CHECK_SUB:
//				autoslab_analyze_minus(gimple_call_arg(def_stmt, 0), gimple_call_arg(def_stmt, 1), c1_out, c0_out);
				break;

			case IFN_UBSAN_CHECK_MUL:
				autoslab_analyze_mult(gimple_call_arg(def_stmt, 0), gimple_call_arg(def_stmt, 1), c1_out, c0_out);
				break;
#endif

#if BUILDING_GCC_VERSION >= 5000
			case IFN_ADD_OVERFLOW:
				autoslab_analyze_plus(gimple_call_arg(def_stmt, 0), gimple_call_arg(def_stmt, 1), c1_out, c0_out);
				break;

			case IFN_MUL_OVERFLOW:
				autoslab_analyze_mult(gimple_call_arg(def_stmt, 0), gimple_call_arg(def_stmt, 1), c1_out, c0_out);
				break;
#endif
			}

			break;
		}

		case GIMPLE_NOP:
			break;

		case GIMPLE_PHI: {
			gphi *phi_stmt;
			unsigned int i, num_args;
			tree origin = NULL_TREE;

			phi_stmt = as_a_gphi(def_stmt);

			for (i = 0, num_args =  gimple_phi_num_args(phi_stmt); i < num_args; i++) {
				tree phi_arg = gimple_phi_arg_def(phi_stmt, i);

				switch (TREE_CODE(phi_arg)) {
				default:
					debug_gimple_stmt(def_stmt);
					debug_tree(phi_arg);
					debug_tree(size);
					gcc_unreachable();

				case COMPLEX_CST:
					phi_arg = TREE_REALPART(phi_arg);
					// FALLTHROUGH

				case INTEGER_CST:
					if (tree_int_cst_lt(KMALLOC_MAX_SIZE, phi_arg))
						continue;
					break;

				case SSA_NAME:
					if (origin == NULL_TREE) {
						origin = phi_arg;
						continue;
					}

					if (origin == phi_arg)
						continue;
					break;
				}

				break;
			}

			if (i < num_args || origin == NULL_TREE)
				break;

			__autoslab_analyze_size(origin, c1_out, c0_out);

			break;
		}
		}
	}
	}
}

static sbitmap autoslab_all_zones(void)
{
	sbitmap gfp_zones;

	gfp_zones = sbitmap_alloc(TREE_INT_CST_LOW(__MAX_NR_ZONES));
	sbitmap_clear(gfp_zones);

	if (ZONE_DMA != NULL_TREE)
		sbitmap_set_bit(gfp_zones, TREE_INT_CST_LOW(ZONE_DMA));
	if (ZONE_DMA32 != NULL_TREE)
		sbitmap_set_bit(gfp_zones, TREE_INT_CST_LOW(ZONE_DMA32));
	sbitmap_set_bit(gfp_zones, TREE_INT_CST_LOW(ZONE_NORMAL));

	return gfp_zones;
}

static sbitmap autoslab_extract_flags(tree flags)
{
	sbitmap gfp_zones;
	unsigned HOST_WIDE_INT flags_val;

	flags_val = TREE_INT_CST_LOW(flags);

	gfp_zones = sbitmap_alloc(TREE_INT_CST_LOW(__MAX_NR_ZONES));
	sbitmap_clear(gfp_zones);

	if (flags_val & tree_to_uhwi(__GFP_DMA) && ZONE_DMA != NULL_TREE)
		sbitmap_set_bit(gfp_zones, TREE_INT_CST_LOW(ZONE_DMA));
	else if (flags_val & tree_to_uhwi(__GFP_DMA32) && ZONE_DMA32 != NULL_TREE)
		sbitmap_set_bit(gfp_zones, TREE_INT_CST_LOW(ZONE_DMA32));
	else
		sbitmap_set_bit(gfp_zones, TREE_INT_CST_LOW(ZONE_NORMAL));

	return gfp_zones;
}

static sbitmap __autoslab_analyze_flags(tree flags, gimple_set *visited, bool *atomic)
{
	switch (TREE_CODE(flags)) {
	default:
		debug_tree(flags);
		gcc_unreachable();

	case INTEGER_CST: {
		unsigned HOST_WIDE_INT flags_val;

		flags_val = TREE_INT_CST_LOW(flags);

		// bail out on __GFP_HIGHMEM/__GFP_MOVABLE
		if (flags_val & (tree_to_uhwi(__GFP_HIGHMEM) | tree_to_uhwi(__GFP_MOVABLE)))
			return NULL;
		// bail out on more than a single zone
		if (!!(flags_val & tree_to_uhwi(__GFP_DMA)) + !!(flags_val & tree_to_uhwi(__GFP_DMA32)) + !!(flags_val & tree_to_uhwi(__GFP_RECLAIMABLE)) > 1)
			return NULL;

		/*
		 * KSTACKOVERFLOW has no problem with these converted GFP_ATOMIC allocations,
		 * and we've disabled the informational message in convert_autostack()
		 */
		//if (flags_val & (tree_to_uhwi(__GFP_ATOMIC)))
		//	*atomic = true;

		return autoslab_extract_flags(flags);
	}

	case SSA_NAME: {
		gimple_stmt_ptr def_stmt;

		def_stmt = SSA_NAME_DEF_STMT(flags);

		switch (gimple_code(def_stmt)) {
		default:
			debug_gimple_stmt(def_stmt);
			debug_tree(flags);
			gcc_unreachable();

		case GIMPLE_ASSIGN: {
			sbitmap gfp_zones;
			tree rhs1;

			rhs1 = gimple_assign_rhs1(def_stmt);

			switch (gimple_assign_rhs_code(def_stmt)) {
			default:
				inform(gimple_expansion_point_location(def_stmt), "code %s", get_tree_code_name(gimple_assign_rhs_code(def_stmt)));
				debug_gimple_stmt(def_stmt);
				debug_tree(flags);
				gcc_unreachable();

			case BIT_AND_EXPR: {
				tree rhs2;
				sbitmap rhs_gfp_zones;

				// shortcut const masks
				gcc_assert(TREE_CODE(rhs1) != INTEGER_CST);
				rhs2 = gimple_assign_rhs2(def_stmt);
				if (TREE_CODE(rhs2) == INTEGER_CST)
					return autoslab_extract_flags(rhs2);

				gfp_zones = __autoslab_analyze_flags(rhs2, visited, atomic);
				if (!gfp_zones)
					return NULL;

				rhs_gfp_zones = __autoslab_analyze_flags(rhs1, visited, atomic);
				if (!rhs_gfp_zones) {
					sbitmap_free(gfp_zones);
					return NULL;
				}
				sbitmap_and(gfp_zones, gfp_zones, rhs_gfp_zones);
				sbitmap_free(rhs_gfp_zones);

				return gfp_zones;
			}

			case COMPONENT_REF:
#if BUILDING_GCC_VERSION >= 4006
			case MEM_REF:
#endif
			case RSHIFT_EXPR:
				// in IPA/LTO we can refine this
				return autoslab_all_zones();

			case COND_EXPR:
#if BUILDING_GCC_VERSION < 4006
				gcc_unreachable();
#endif
			case BIT_IOR_EXPR: {
				sbitmap rhs_gfp_zones;

				gfp_zones = __autoslab_analyze_flags(gimple_assign_rhs2(def_stmt), visited, atomic);
				if (!gfp_zones)
					return NULL;

#if BUILDING_GCC_VERSION < 4006
				rhs_gfp_zones = __autoslab_analyze_flags(rhs1, visited, atomic);
#else
				rhs_gfp_zones = __autoslab_analyze_flags(gimple_assign_rhs_code(def_stmt) == BIT_IOR_EXPR ? rhs1 : gimple_assign_rhs3(def_stmt), visited, atomic);
#endif
				if (!rhs_gfp_zones) {
					sbitmap_free(gfp_zones);
					return NULL;
				}
				sbitmap_ior(gfp_zones, gfp_zones, rhs_gfp_zones);
				sbitmap_free(rhs_gfp_zones);

				return gfp_zones;
			}

			case EQ_EXPR:
			case GE_EXPR:
			case GT_EXPR:
			case LE_EXPR:
			case LT_EXPR:
			case NE_EXPR:
				return autoslab_extract_flags(integer_one_node);

			case PLUS_EXPR: {
				sbitmap rhs_gfp_zones;

				gfp_zones = __autoslab_analyze_flags(gimple_assign_rhs2(def_stmt), visited, atomic);
				if (!gfp_zones)
					return NULL;

				rhs_gfp_zones = __autoslab_analyze_flags(rhs1, visited, atomic);
				if (!rhs_gfp_zones) {
					sbitmap_free(gfp_zones);
					return NULL;
				}
				sbitmap_ior(gfp_zones, gfp_zones, rhs_gfp_zones);
				sbitmap_free(rhs_gfp_zones);

				return gfp_zones;
			}

			case NOP_EXPR:
			case VAR_DECL:
				return __autoslab_analyze_flags(rhs1, visited, atomic);
			}
		}

		case GIMPLE_CALL:
			// in IPA/LTO we can refine this
			return autoslab_all_zones();

		case GIMPLE_NOP: {
			sbitmap gfp_zones;

			// in IPA/LTO we can refine this
			gfp_zones = autoslab_all_zones();

#if BUILDING_GCC_VERSION >= 7000
			tree _value;
			widest_int _mask;
			unsigned HOST_WIDE_INT value, mask;
			location_t loc = DECL_SOURCE_LOCATION(current_function_decl);

			gcc_assert(SSA_NAME_VAR(flags));
			flags = SSA_NAME_VAR(flags);
			gcc_assert(TREE_CODE(flags) == PARM_DECL);

			if (!ipcp_get_parm_bits(flags, &_value, &_mask))
				return gfp_zones;

			value = tree_to_uhwi(_value);
			mask = _mask.to_uhwi();

			if (ZONE_DMA != NULL_TREE && !(mask & tree_to_uhwi(__GFP_DMA)) && !(value & tree_to_uhwi(__GFP_DMA)))
				sbitmap_clear_bit(gfp_zones, TREE_INT_CST_LOW(ZONE_DMA));
			if (ZONE_DMA32 != NULL_TREE && !(mask & tree_to_uhwi(__GFP_DMA32)) && !(value & tree_to_uhwi(__GFP_DMA32)))
				sbitmap_clear_bit(gfp_zones, TREE_INT_CST_LOW(ZONE_DMA32));

			if (verbose)
				inform(loc, "autoslab: FLAGS %qD VALUE %lx MASK %lx", flags, value, mask);
#endif

			return gfp_zones;
		}

		case GIMPLE_PHI: {
			sbitmap phi_gfp_zones;
			gphi *phi_stmt;
			unsigned int i, num_args;

			phi_gfp_zones = sbitmap_alloc(TREE_INT_CST_LOW(__MAX_NR_ZONES));
			sbitmap_clear(phi_gfp_zones);

			if (pointer_set_insert(visited, def_stmt))
				return phi_gfp_zones;

			phi_stmt = as_a_gphi(def_stmt);

			for (i = 0, num_args = gimple_phi_num_args(phi_stmt); i < num_args; i++) {
				sbitmap arg_gfp_zones;

				arg_gfp_zones = __autoslab_analyze_flags(gimple_phi_arg_def(phi_stmt, i), visited, atomic);
				if (!arg_gfp_zones) {
					sbitmap_free(phi_gfp_zones);
					return NULL;
				}

				sbitmap_ior(phi_gfp_zones, phi_gfp_zones, arg_gfp_zones);
				sbitmap_free(arg_gfp_zones);
			}

			return phi_gfp_zones;
		}
		}
	}

	case VAR_DECL:
		return autoslab_all_zones();
	}
}

static sbitmap autoslab_analyze_flags(gimple_stmt_iterator *gsi, tree flags, bool *atomic)
{
	gimple_set *visited;
	sbitmap gfp_zones;

	visited = gimple_pointer_set_create();
	gfp_zones = __autoslab_analyze_flags(flags, visited, atomic);
	pointer_set_destroy(visited);

	return gfp_zones;
}

/*
 * convert a k*alloc*(size, flags, ...) call into kmem_cache_*alloc*(cache, flags, ...)
 *
 * in particular:
 * 1. kmalloc            -> kmem_cache_alloc
 * 2. kmalloc_node       -> kmem_cache_alloc_node
 * 3. kzalloc            -> kmem_cache_alloc
 * 4. kzalloc_node       -> kmem_cache_alloc_node
 * 5. kcalloc            -> kmem_cache_alloc
 * 6. kmalloc_array      -> kmem_cache_alloc
 * 7. kcalloc_node       -> kmem_cache_alloc_node
 * 8. kmalloc_array_node -> kmem_cache_alloc_node
 */
static void autoslab_convert_call(gimple_stmt_iterator *gsi,
				  unsigned int kalloc_func,
				  tree count_arg,
				  tree size_arg,
				  tree flags_arg,
				  tree node_arg)
{
	gcall *kalloc_call_stmt;
	enum kmem_call_names kmem_call;
	tree lhs, lhs_type, size, flags, node, lhs_size, maxsize;
	unsigned HOST_WIDE_INT size_val, lhs_size_val, lhs_align_val;
	location_t loc;
	const char *lhs_type_name_ptr;
	sbitmap gfp_zones;
	bool atomic;
#if BUILDING_GCC_VERSION >= 4006
	struct ptr_info_def *pi;
	unsigned int misalign;
#endif
	unsigned int align, c1, c0, minshift, maxshift;

	gcc_assert(count_arg == NULL_TREE);

	kalloc_call_stmt = as_a_gcall(gsi_stmt(*gsi));
	loc = gimple_expansion_point_location(kalloc_call_stmt);

	lhs = gimple_call_lhs(kalloc_call_stmt);
	lhs_type = autoslab_get_alloc_type(kalloc_call_stmt);
	gcc_assert(lhs_type != NULL_TREE);

	size = gimple_call_arg(kalloc_call_stmt, TREE_INT_CST_LOW(size_arg) - 1);
	flags = gimple_call_arg(kalloc_call_stmt, TREE_INT_CST_LOW(flags_arg) - 1);
	node = node_arg ? gimple_call_arg(kalloc_call_stmt, TREE_INT_CST_LOW(node_arg) - 1) : NULL_TREE;

	lhs_type_name_ptr = "";
	if (RECORD_OR_UNION_TYPE_P(lhs_type)) {
		const_tree lhs_type_name;

		lhs_type_name = type_name(lhs_type);
		if (lhs_type_name)
			lhs_type_name_ptr = IDENTIFIER_POINTER(lhs_type_name);
	}

	lhs_size = TYPE_SIZE_UNIT(lhs_type);
	if (lhs_size != NULL_TREE) {
		gcc_assert(TREE_CODE(lhs_size) == INTEGER_CST);
		lhs_size_val = TREE_INT_CST_LOW(lhs_size);
	} else
		lhs_size_val = 0;

	atomic = false;
	gfp_zones = autoslab_analyze_flags(gsi, flags, &atomic);

	if (!gfp_zones) {
		error_at(loc, "autoslab: no badflags size: %qE lhssize:%qE flags:%qE lhstype:%qs",
			 size, lhs_size ?: null_pointer_node, flags, lhs_type_name_ptr);
		return;
	}

	gcc_assert(sbitmap_count_bits(gfp_zones) > 0);

	maxsize = autoslab_analyze_size(gsi, size, &c1, &c0, &minshift, &maxshift);

	// we managed to reduce varsize to constsize
	if (minshift == maxshift)
		size = maxsize;
	size_val = TREE_INT_CST_LOW(maxsize);
	if (lhs_size_val == 0)
		lhs_size_val = size_val;

	// handle the special sizes that return a const ptr
	if (integer_zerop(size)) {
		// convert 0-sized allocations to ZERO_SIZE_PTR
		gcc_assert(!SSA_NAME_OCCURS_IN_ABNORMAL_PHI(lhs));
		replace_call_with_value(gsi, ZERO_SIZE_PTR);

		if (verbose)
			inform(loc, "autoslab: yes nullsize %qE %qs", flags, lhs_type_name_ptr);
		goto out;
	}

	if (TREE_CODE(size) == INTEGER_CST && tree_int_cst_lt(KMALLOC_MAX_SIZE, size)) {
		// convert oversized allocations to NULL
		gcc_assert(!SSA_NAME_OCCURS_IN_ABNORMAL_PHI(lhs));
		replace_call_with_value(gsi, null_pointer_node);

		if (verbose)
			inform(loc, "autoslab: yes maxsize "
				HOST_WIDE_INT_PRINT_UNSIGNED "/" HOST_WIDE_INT_PRINT_HEX_PURE
				" %qE %qs",
				size_val, size_val, flags, lhs_type_name_ptr);
		goto out;
	}

	// compute the required alignment
	lhs_align_val = MAX(TYPE_ALIGN_UNIT(lhs_type), TREE_INT_CST_LOW(ARCH_KMALLOC_MINALIGN));
#if BUILDING_GCC_VERSION == 4005
	align = get_pointer_alignment(lhs, BIGGEST_ALIGNMENT);
#else
	pi = get_ptr_info(lhs);
	if (!get_ptr_info_alignment(pi, &align, &misalign))
		align = lhs_align_val;
#endif
	lhs_align_val = MAX(lhs_align_val, align);

	// try to convert to an automatic variable first
	if (can_convert_to_automatic(gsi, maxsize, gfp_zones, atomic, lhs_align_val)) {
		autoslab_convert_to_automatic(gsi, maxsize, flags, gfp_zones, lhs_size_val, lhs_align_val);

		if (verbose)
			inform(loc, "autoslab: yes stack "
				    HOST_WIDE_INT_PRINT_UNSIGNED "/" HOST_WIDE_INT_PRINT_HEX_PURE " "
				    HOST_WIDE_INT_PRINT_UNSIGNED "/" HOST_WIDE_INT_PRINT_HEX_PURE " "
				    "%qE %s %lu %u %qT",
				lhs_size_val, lhs_size_val,
				size_val, size_val,
				flags, node ? "node" : "nonode",
				lhs_align_val,
				TYPE_ALIGN_UNIT(TREE_TYPE(TREE_TYPE(lhs))),
				TYPE_MAIN_VARIANT(TREE_TYPE(TREE_TYPE(lhs))));
	}

	align = MIN(1U << ceil_log2(size_val), tree_to_uhwi(SMP_CACHE_BYTES));
	if (TREE_CODE(size) == INTEGER_CST)
		align = MIN(1U << ctz_hwi(size_val), align);
	else
		align = MIN(1U << ceil_log2(size_val), align);
	lhs_align_val = MAX(lhs_align_val, align);

	if (TREE_CODE(size) != INTEGER_CST) {
		tree kmem_caches;

		if (sbitmap_count_bits(gfp_zones) == 1)
			// build varsize:
			// ptr = kmem_cache_alloc_index(&autoslabs, size, flags, minshift, maxshift);
			kmem_call = kalloc_functions[kalloc_func].kmem_name_index;
		else
			// build varsizeflags:
			// ptr = kmem_cache_alloc_index_type(&autoslabs, size, flags, minshift, maxshift);
			kmem_call = kalloc_functions[kalloc_func].kmem_name_index_type;

		if (kmem_cache_decls[kmem_call].decl == NULL_TREE) {
			error_at(loc, "autoslab did not find %qs to replace %qD", kmem_cache_decls[kmem_call].name, gimple_call_fndecl(kalloc_call_stmt));
			goto out;
		}

		gcc_assert(minshift < maxshift);
		kmem_caches = autoslab_create_infos_size(gsi, size, flags, gfp_zones, lhs_type_name_ptr, c1, c0, minshift, maxshift, size_val, lhs_align_val);
		kmem_caches = build1_loc(loc, ADDR_EXPR, TYPE_POINTER_TO(autoslab_infos_type), kmem_caches);
		kmem_caches = fold_convert_loc(loc, TYPE_POINTER_TO(TYPE_POINTER_TO(autoslab_info_type)), kmem_caches);
		__autoslab_convert_call(gsi, kmem_call, size, flags, gfp_zones, node, build_int_cstu(unsigned_type_node, minshift), build_int_cstu(unsigned_type_node, maxshift), lhs_type_name_ptr, lhs_size_val, lhs_align_val, kmem_caches);
	} else if (sbitmap_count_bits(gfp_zones) == 1) {
		tree autoslab_cache_id, kmem_cache, autoslab_info, ctor;
		char *autoslab_cache_name;
		varpool_node_ptr var;
		unsigned HOST_WIDE_INT flags_val;

		flags_val = TREE_CODE(flags) == INTEGER_CST ? TREE_INT_CST_LOW(flags) : 0;

		// build constsize:
		// ptr = kmem_cache_alloc(autoslab, flags);
		kmem_call = kalloc_functions[kalloc_func].kmem_name;
		if (kmem_cache_decls[kmem_call].decl == NULL_TREE) {
			error_at(loc, "autoslab did not find %qs to replace %qD", kmem_cache_decls[kmem_call].name, gimple_call_fndecl(kalloc_call_stmt));
			goto out;
		}

		gcc_assert(-1 != autoslab_gen_name(&autoslab_cache_name, "cache_", "", gsi, size_val, lhs_align_val, flags_val, gfp_zones, 0, 0, 0, 0));
		autoslab_cache_id = get_identifier(autoslab_cache_name);
		free(autoslab_cache_name);
		var = varpool_get_for_asmname(autoslab_cache_id);
		gcc_assert(!var);

		autoslab_info = autoslab_create_info(gsi, size, flags, gfp_zones, "const_", lhs_type_name_ptr, lhs_align_val);
		gcc_assert(TYPE_POINTER_TO(autoslab_info_type));
		gcc_assert(POINTER_TYPE_P(TYPE_POINTER_TO(autoslab_info_type)));
		ctor = build1_loc(loc, ADDR_EXPR, TYPE_POINTER_TO(autoslab_info_type), autoslab_info);
		kmem_cache = autoslab_build_cache_ptr(loc, autoslab_cache_id, TREE_VALUE(TYPE_ARG_TYPES(TREE_TYPE(kmem_cache_decls[kmem_call].decl))), ctor);
		__autoslab_convert_call(gsi, kmem_call, size, flags, gfp_zones, node, NULL_TREE, NULL_TREE, lhs_type_name_ptr, lhs_size_val, lhs_align_val, kmem_cache);
	} else {
		tree kmem_caches;

		gcc_assert(TREE_CODE(flags) == SSA_NAME);

		// build varflags:
		// ptr = kmem_cache_alloc_type(&autoslabs, flags);
		kmem_call = kalloc_functions[kalloc_func].kmem_name_type;
		if (kmem_cache_decls[kmem_call].decl == NULL_TREE) {
			error_at(loc, "autoslab did not find %qs to replace %qD", kmem_cache_decls[kmem_call].name, gimple_call_fndecl(kalloc_call_stmt));
			goto out;
		}

		kmem_caches = autoslab_create_infos_type(gsi, size, gfp_zones, lhs_type_name_ptr, lhs_align_val);
		kmem_caches = build1_loc(loc, ADDR_EXPR, TYPE_POINTER_TO(autoslab_infos_type), kmem_caches);
		kmem_caches = fold_convert_loc(loc, TYPE_POINTER_TO(TYPE_POINTER_TO(autoslab_info_type)), kmem_caches);
		__autoslab_convert_call(gsi, kmem_call, size, flags, gfp_zones, node, NULL_TREE, NULL_TREE, lhs_type_name_ptr, lhs_size_val, lhs_align_val, kmem_caches);
	}

out:
	sbitmap_free(gfp_zones);
}

static unsigned int autoslab_execute(void)
{
	basic_block bb;

	loop_optimizer_init(LOOPS_NORMAL | LOOPS_HAVE_RECORDED_EXITS);
	gcc_assert(current_loops);

	calculate_dominance_info(CDI_DOMINATORS);
	calculate_dominance_info(CDI_POST_DOMINATORS);

	gcc_assert(dom_info_available_p(CDI_DOMINATORS));
	gcc_assert(dom_info_available_p(CDI_POST_DOMINATORS));

#if BUILDING_GCC_VERSION >= 4009
	scev_initialize();
#endif

	find_unreachable_blocks();

#if BUILDING_GCC_VERSION >= 8000 && BUILDING_GCC_VERSION < 9000
	if (!ipa_call_summaries)
		ipa_call_summaries = new ipa_call_summary_t(symtab, false);
#endif

	// 1. loop through BBs and GIMPLE statements
	FOR_EACH_BB_FN(bb, cfun) {
		gimple_stmt_iterator gsi;

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			tree fndecl, count_arg, size_arg, flags_arg, node_arg;
			gimple_stmt_ptr stmt;
			gcall *call_stmt;
			unsigned int kalloc_func;

			// 2. is it a call ...
			stmt = gsi_stmt(gsi);
			if (!is_gimple_call(stmt))
				continue;
			call_stmt = as_a_gcall(stmt);
			fndecl = gimple_call_fndecl(call_stmt);
			if (fndecl == NULL_TREE)
				continue;
			if (TREE_CODE(fndecl) != FUNCTION_DECL)
				continue;

			// 3. ... to k*alloc
			kalloc_func = is_kalloc(DECL_NAME_POINTER(fndecl));
			if (kalloc_func >= ARRAY_SIZE(kalloc_functions))
				continue;

			// 4. remove stray kmalloc calls left behind by certain transformations
			if (gimple_call_lhs(call_stmt) == NULL_TREE) {
				unlink_stmt_vdef(call_stmt);
				gsi_remove(&gsi, true);
				if (gsi_end_p(gsi))
					break;
				gsi_prev(&gsi);
				continue;
			}

			// 5. the alloc/malloc/alloc_size/alloc_flags attrs must be present on the decl/type
			gcc_assert(lookup_attribute("alloc", DECL_ATTRIBUTES(fndecl)));
			gcc_assert(lookup_attribute("malloc", DECL_ATTRIBUTES(fndecl)));
			autoslab_get_alloc_attrs(fndecl, &count_arg, &size_arg, &flags_arg, NULL, &node_arg);
			gcc_assert(flags_arg);

			// 6. try to convert the allocation into a type specific slab allocation
			autoslab_convert_call(&gsi, kalloc_func, count_arg, size_arg, flags_arg, node_arg);
		}
	}

#if BUILDING_GCC_VERSION >= 4009
	scev_finalize();
#endif

	free_dominance_info(CDI_POST_DOMINATORS);
	free_dominance_info(CDI_DOMINATORS);
	loop_optimizer_finalize();

	return 0;
}

#define PASS_NAME autoslab
#define NO_GATE
#define PROPERTIES_REQUIRED PROP_cfg
#define TODO_FLAGS_START TODO_rebuild_cgraph_edges
#define TODO_FLAGS_FINISH TODO_verify_ssa | TODO_verify_stmts | TODO_dump_func | TODO_remove_unused_locals | TODO_update_ssa | TODO_cleanup_cfg | TODO_ggc_collect | TODO_rebuild_cgraph_edges | TODO_verify_flow | TODO_update_address_taken
#include "gcc-generate-gimple-pass.h"
