#ifndef RAP_H_INCLUDED
#define RAP_H_INCLUDED

#include "gcc-common.h"
#include <regex.h>

typedef struct {
	int hash; // will be sign extended to long in reality
} rap_hash_t;

typedef struct {
	unsigned int qual_const:1;
	unsigned int qual_volatile:1;
	unsigned int func_name:1;
	unsigned int func_scope:1;
} rap_hash_flags_t;
extern rap_hash_flags_t precise_rap_hash_flags;
extern rap_hash_flags_t imprecise_rap_hash_flags;

extern bool report_fptr_hash;
extern bitmap complex_functions;
extern GTY(()) tree rap_cookie_decl;
#define rap_hash_type_node long_integer_type_node
#ifdef TARGET_386
extern GTY(()) rtx leaveq_rtx;
extern GTY(()) rtx pop_rbp_rtx;
#endif
extern unsigned int rap_reg_id;
extern const char *rap_reg_name;
extern const char *asm_call_insn_pattern;
extern regex_t asm_call_insn_regex;
extern const char *rap_ret_abort;
extern const char *rap_call_abort;
extern bool rap_include, enable_xor, enable_type_ret, enable_type_call, enable_type_nospec;

extern void (*kernexec_instrument_fptr)(gimple_stmt_iterator *);
extern void (*kernexec_instrument_retaddr)(rtx_insn *);
void kernexec_instrument_fptr_bts(gimple_stmt_iterator *gsi);
void kernexec_instrument_fptr_or(gimple_stmt_iterator *gsi);
void kernexec_instrument_retaddr_bts(rtx_insn *insn);
void kernexec_instrument_retaddr_or(rtx_insn *insn);

void siphash24fold(unsigned char *out, const unsigned char *in, unsigned long long inlen, const unsigned char *k);
void rap_calculate_func_hashes(void *event_data, void *data);
void rap_calculate_fptr_hashes(void *event_data, void *data);
void rap_hash_compute_attr(void *event_data, void *data __unused);
void rap_calculate_fndecl_hash(tree fndecl);

extern unsigned int rap_xor_checks[2];
void rap_all_passes_start(void *event_data __unused, void *data __unused);

rap_hash_t rap_lookup_imprecise_rap_hash(const_tree fntype);
tree get_rap_hash(gimple_seq *stmts, location_t loc, tree fptr, HOST_WIDE_INT rap_hash_offset);
const_tree type_name(const_tree type);

gimple_stmt_ptr ibarrier(tree var);
bool rap_cmodel_check(void);
bool rap_ret_gate(void);
bool rap_is_complex_function(tree fndecl);
bool rap_noreturn_function(void);
enum rap_ret_gate_ignore_calls { ignore_none, ignore_tail, ignore_all };
bool rap_ret_gate_bb(basic_block bb, enum rap_ret_gate_ignore_calls);
bool rap_calls_complex_function_p(tree fndecl);
void check_register_variable(tree var);
void rap_hash_finish_type(void *event_data, void *data);
void rap_mark_retloc(rtx_insn *insn, rtx_insn *reference);
bool rap_retpoline_gate(void);

#ifdef TARGET_386
#define ADDR_SPACE_KERNEXEC	4

static inline rtx gen_pop(rtx arg)
{
	return gen_rtx_set(arg, gen_rtx_MEM(Pmode, gen_rtx_POST_INC(Pmode, stack_pointer_rtx)));
}
#endif

#if BUILDING_GCC_VERSION >= 4009
ipa_opt_pass_d *make_pass_rap_ret_ipa(gcc::context *);
gimple_opt_pass *make_pass_rap_check_tail_calls(gcc::context *);
gimple_opt_pass *make_pass_rap_ret(gcc::context *);
gimple_opt_pass *make_pass_rap_reload(gcc::context *);
gimple_opt_pass *make_pass_rap_fptr(gcc::context *);
gimple_opt_pass *make_pass_rap_check_local_register_variables(gcc::context *);
rtl_opt_pass *make_pass_rap_xor_elimination(gcc::context *);
rtl_opt_pass *make_pass_rap_rewrite_retaddr_check(gcc::context *);
rtl_opt_pass *make_pass_rap_fptr_safety(gcc::context *);
rtl_opt_pass *make_pass_rap_mark_retloc(gcc::context *);
rtl_opt_pass *make_pass_rap_retpoline(gcc::context *);
rtl_opt_pass *make_pass_rap_indirect_branch_register(gcc::context *);
gimple_opt_pass *make_pass_kernexec_reload(gcc::context *);
gimple_opt_pass *make_pass_kernexec_fptr(gcc::context *);
rtl_opt_pass *make_pass_kernexec_retaddr(gcc::context *);
#else
struct opt_pass *make_pass_rap_ret_ipa(void);
struct opt_pass *make_pass_rap_check_tail_calls(void);
struct opt_pass *make_pass_rap_ret(void);
struct opt_pass *make_pass_rap_reload(void);
struct opt_pass *make_pass_rap_fptr(void);
struct opt_pass *make_pass_rap_check_local_register_variables(void);
struct opt_pass *make_pass_rap_xor_elimination(void);
struct opt_pass *make_pass_rap_rewrite_retaddr_check(void);
struct opt_pass *make_pass_rap_fptr_safety(void);
struct opt_pass *make_pass_rap_mark_retloc(void);
struct opt_pass *make_pass_rap_retpoline(void);
struct opt_pass *make_pass_rap_indirect_branch_register(void);
struct opt_pass *make_pass_kernexec_reload(void);
struct opt_pass *make_pass_kernexec_fptr(void);
struct opt_pass *make_pass_kernexec_retaddr(void);
static inline struct opt_pass *make_pass_merge_phi(void) { return &pass_merge_phi.pass; };
#endif

#endif
